import logging
import asyncio
import re
import csv
import io
import uuid
from datetime import datetime
from typing import Dict, List, Optional, Any, Union, Tuple, AsyncGenerator, Set
from contextlib import asynccontextmanager
import psycopg
from psycopg import sql
from psycopg.rows import dict_row

# RDFLib imports for term handling
from rdflib import URIRef, Literal, BNode, Variable
from rdflib.term import Identifier

# Import space backend interface
from ..space_backend_interface import SpaceBackendInterface, SparqlBackendInterface

# Import PostgreSQL utilities
from .postgresql_log_utils import PostgreSQLLogUtils
from .space.postgresql_space_utils import PostgreSQLSpaceUtils
from .postgresql_cache_term import PostgreSQLCacheTerm
from .postgresql_cache_graph import PostgreSQLCacheGraph
from .space.postgresql_space_core import PostgreSQLSpaceCore
from .space.postgresql_space_schema import PostgreSQLSpaceSchema
from .space.postgresql_space_db_mgmt import PostgreSQLSpaceDbMgmt
from .space.postgresql_space_namespaces import PostgreSQLSpaceNamespaces
from .space.postgresql_space_datatypes import PostgreSQLSpaceDatatypes
from .space.postgresql_space_terms import PostgreSQLSpaceTerms
from .space.postgresql_space_queries import PostgreSQLSpaceQueries
from .space.postgresql_space_db_ops import PostgreSQLSpaceDBOps
from .space.postgresql_space_db_objects import PostgreSQLSpaceDBObjects
from .space.postgresql_space_graphs import PostgreSQLSpaceGraphs
from .space.postgresql_space_db_import import PostgreSQLSpaceDBImport

class PostgreSQLSpaceImpl(SpaceBackendInterface):
    """
    PostgreSQL implementation for RDF graph space management.
    
    Manages all RDF spaces and their associated tables:
    - namespace: RDF namespace prefix mappings per space
    - graph: Graph (context) URIs and metadata per space
    - rdf_quad: RDF quads with term references per space
    - term: Terms dictionary with types and metadata per space
    
    Table names are prefixed with global_prefix__space_id__ format.
    """
    
    def __init__(self, connection_string: str, global_prefix: str = "vitalgraph", pool_config: Optional[Dict[str, Any]] = None, shared_pool=None, dict_pool=None, db_impl=None):
        """
        Initialize PostgreSQL space implementation with shared or dedicated psycopg3 ConnectionPool.
        
        Args:
            connection_string: PostgreSQL connection string
            global_prefix: Global prefix for table names
            pool_config: Optional connection pool configuration
            shared_pool: Optional shared psycopg3 ConnectionPool instance (for tuple results)
            dict_pool: Optional dict psycopg3 ConnectionPool instance (for dictionary results)
            db_impl: Database implementation instance for signal manager access
        """
        self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
        
        # Initialize core connection management
        self.core = PostgreSQLSpaceCore(
            db_impl=db_impl,
            connection_string=connection_string,
            global_prefix=global_prefix,
            pool_config=pool_config,
            shared_pool=shared_pool,
            dict_pool=dict_pool
        )
        
        # Store db_impl reference for signal manager access
        self.db_impl = db_impl
        
        # Keep these attributes for backward compatibility and space-specific functionality
        self.connection_string = connection_string
        self.global_prefix = global_prefix
        
        
        # Validate global prefix using utils
        PostgreSQLSpaceUtils.validate_global_prefix(global_prefix)
        
        # Initialize database management class
        self.db_mgmt = PostgreSQLSpaceDbMgmt(self)
        
        # Initialize namespace management class
        self.namespaces = PostgreSQLSpaceNamespaces(self)
        
        # Initialize datatype management class
        self.datatypes = PostgreSQLSpaceDatatypes(self)
        
        # Initialize term management class
        self.terms = PostgreSQLSpaceTerms(self)
        
        # Initialize query management class
        self.queries = PostgreSQLSpaceQueries(self)
        
        # Initialize database operations class
        self.db_ops = PostgreSQLSpaceDBOps(self)
        
        # Initialize database objects class
        self.db_objects = PostgreSQLSpaceDBObjects(self)
        
        # Initialize graph management class
        self.graphs = PostgreSQLSpaceGraphs(self)
        
        # Initialize import management class
        self.db_import = PostgreSQLSpaceDBImport(self)
        
        # Initialize PostgreSQL utilities
        self.utils = PostgreSQLSpaceUtils()
        
        # Cache of schema instances by space_id
        self._schema_cache = {}
        
        # Cache of SPARQL implementations by space_id to persist term caches across requests
        self._sparql_impl_cache = {}
        
        # Cache of datatype caches by space_id - each space has its own datatype cache
        self._datatype_caches = {}
        
        # Cache of graph caches by space_id - each space has its own graph cache
        self._graph_caches = {}
        
        # Global term cache for all spaces (shared for performance)
        self._term_cache = PostgreSQLCacheTerm()
        
        # Flag to track if datatype cache has been loaded
        self._datatype_cache_loaded = False
    
    def _get_schema(self, space_id: str) -> PostgreSQLSpaceSchema:
        """
        Get or create a schema instance for the specified space.
        
        Args:
            space_id: Space identifier
            
        Returns:
            PostgreSQLSpaceSchema: Schema instance for the space
        """
        if space_id not in self._schema_cache:
            self._schema_cache[space_id] = PostgreSQLSpaceSchema(
                global_prefix=self.global_prefix,
                space_id=space_id
            )
        return self._schema_cache[space_id]
    
    def get_signal_manager(self):
        """
        Get access to the SignalManager instance.
        
        This method provides access to the SignalManager to child components
        like PostgreSQLSpaceGraphs without creating circular imports.
        
        Returns:
            SignalManager: The application-wide SignalManager instance or None if not initialized
        """
        return self.db_impl.get_signal_manager()

    def get_term_cache(self):
        """
        Get the term cache for this PostgreSQL space implementation.
        
        Returns:
            PostgreSQLCacheTerm: The term cache instance
        """
        return self._term_cache
    
    def get_manager_info(self) -> Dict[str, Any]:
        """
        Get general information about this space manager.
        
        Returns:
            dict: Manager information
        """
        return self.core.get_manager_info(self)

    def get_datatype_cache(self, space_id: str) -> 'PostgreSQLCacheDatatype':
        """
        Get the datatype cache for a specific space, creating it if necessary.
        Each space has its own datatype cache that is initialized when the space is created.
        
        Args:
            space_id: Space identifier
            
        Returns:
            PostgreSQLCacheDatatype: The datatype cache for the specified space
        """
        return self.datatypes.get_datatype_cache(space_id)
    
    def get_graph_cache(self, space_id: str) -> 'PostgreSQLCacheGraph':
        """
        Get the graph cache for a specific space, creating it if necessary.
        Each space has its own graph cache for tracking graph existence.
        The cache is automatically initialized with all existing graphs from the database.
        
        Args:
            space_id: Space identifier
            
        Returns:
            PostgreSQLCacheGraph: The graph cache for the specified space
        """
        if space_id not in self._graph_caches:
            # Create new cache (will be lazily initialized on first async access)
            cache = PostgreSQLCacheGraph()
            self.logger.info(f"Created graph cache for space '{space_id}' (will be initialized on first access)")
            
            self._graph_caches[space_id] = cache
            
        return self._graph_caches[space_id]
    
    def close(self):
        """
        Close the RDF connection pool and clean up resources.
        """
        self.core.close()
    
    def __del__(self):
        """
        Destructor to ensure pool is closed when object is garbage collected.
        """
        try:
            self.close()
        except Exception:
            pass  # Ignore errors during cleanup
      
    @asynccontextmanager
    async def get_db_connection(self):
        """
        Async context manager for automatic connection management using shared or dedicated psycopg3 pool.
        
        Usage:
            async with self.get_db_connection() as conn:
                # Use connection
                cursor = conn.cursor()
                # ... database operations
        """
        async with self.core.get_db_connection() as conn:
            yield conn
    
    def _resolve_term_info(self, term: Identifier) -> Tuple[str, str, Optional[str], Optional[int]]:
        """
        Resolve an RDF term to its database representation.
        
        Args:
            term: RDFLib term (URIRef, Literal, BNode)
            
        Returns:
            tuple: (term_text, term_type, lang, datatype_id)
        """
        return self.terms._resolve_term_info(term)
    
    def _get_table_names(self, space_id: str) -> Dict[str, str]:
        """
        Get all RDF space table names for a specific space.
        
        Args:
            space_id: Space identifier
            
        Returns:
            dict: Dictionary of table names keyed by base name
        """
        schema = self._get_schema(space_id)
        return schema.get_table_names()
    
    def get_sparql_impl(self, space_id: str):
        """
        Get or create a cached SPARQL implementation for the given space.
        
        This ensures that the SPARQL implementation and its term cache persist
        across multiple requests, providing significant performance benefits.
        
        Args:
            space_id: Space identifier
            
        Returns:
            PostgreSQLSparqlImpl: Cached SPARQL implementation instance
        """
        if space_id not in self._sparql_impl_cache:
            # Import here to avoid circular imports
            from .postgresql_sparql_impl import PostgreSQLSparqlImpl
            
            self.logger.info(f" Creating new cached SPARQL implementation for space: {space_id}")
            self._sparql_impl_cache[space_id] = PostgreSQLSparqlImpl(self)
        else:
            self.logger.debug(f" Reusing cached SPARQL implementation for space: {space_id}")
            
        return self._sparql_impl_cache[space_id]
    
    def _get_create_table_sql(self, space_id: str) -> Dict[str, str]:
        """
        Generate CREATE TABLE SQL statements for UUID-based RDF space tables.
        
        This creates tables optimized for UUID-based term identification:
        - Uses UUID primary keys instead of BIGSERIAL
        - Eliminates foreign key constraints for better performance
        - Optimized for deterministic UUID-based batch loading
        
        Args:
            space_id: Space identifier
            
        Returns:
            dict: Dictionary of CREATE TABLE SQL statements
        """
        schema = self._get_schema(space_id)
        return schema.get_create_table_sql()
    
    def _get_standard_datatypes(self) -> List[Tuple[str, str]]:
        """
        Get list of standard XSD and RDF datatype URIs to insert into datatype table.
        
        Returns:
            List of (datatype_uri, datatype_name) tuples
        """
        return self.datatypes.get_standard_datatypes()
    
    def _insert_standard_datatypes(self, space_id: str) -> bool:
        """
        Insert standard datatype URIs into the datatype table for a space.
        
        Args:
            space_id: Space identifier
            
        Returns:
            bool: True if successful, False otherwise
        """
        return self.datatypes.insert_standard_datatypes(space_id)
    
    async def _ensure_text_search_extensions(self) -> bool:
        """
        Ensure that required PostgreSQL extensions for text search are enabled.
        This includes pg_trgm for trigram-based regex matching.
        
        Returns:
            bool: True if extensions are available, False otherwise
        """
        return await self.core._ensure_text_search_extensions()
    
    async def _ensure_datatype_cache_loaded(self, space_id: str) -> None:
        """
        Ensure datatype cache is loaded for a specific space.
        This method loads datatypes from the database into the shared cache if not already loaded.
        
        Args:
            space_id: Space identifier
        """
        return await self.datatypes.ensure_datatype_cache_loaded(space_id)
    
    async def _load_all_datatypes_into_cache(self) -> None:
        """
        Load all datatypes from all spaces into the shared datatype cache.
        This is more efficient than loading per-space since datatypes are shared.
        """
        return await self.datatypes.load_all_datatypes_into_cache()
    
    async def load_datatype_cache(self, space_id: str) -> 'PostgreSQLCacheDatatype':
        """
        Get the datatype cache for a specific space.
        This method ensures the cache is loaded and returns the shared instance.
        
        Args:
            space_id: Space identifier
            
        Returns:
            PostgreSQLCacheDatatype instance populated with datatypes
        """
        return await self.datatypes.load_datatype_cache(space_id)
    
    async def get_or_create_datatype_id(self, space_id: str, datatype_uri: str) -> int:
        """
        Get or create a datatype ID for the given URI.
        This method first checks the cache, then the database, and creates a new entry if needed.
        
        Args:
            space_id: Space identifier
            datatype_uri: The datatype URI to resolve
            
        Returns:
            The datatype ID (BIGINT)
        """
        return await self.datatypes.get_or_create_datatype_id(space_id, datatype_uri)
    
    async def _resolve_datatype_ids_batch(self, space_id: str, datatype_uris: set) -> Dict[str, int]:
        """
        Resolve datatype URIs to IDs, inserting unknown datatypes as needed.
        
        Args:
            space_id: Space identifier for the datatype table
            datatype_uris: Set of datatype URIs to resolve
            
        Returns:
            Dictionary mapping datatype URIs to their BIGINT IDs
        """
        return await self.datatypes.resolve_datatype_ids_batch(space_id, datatype_uris)
    
    async def _process_term_with_datatype(self, space_id: str, term_value: str, term_type: str, lang: Optional[str], datatype_uri: Optional[str]) -> Tuple[str, str, Optional[str], Optional[int]]:
        """
        Helper method to process a term and resolve its datatype ID using the space-specific cache.
        
        Args:
            space_id: Space identifier for the datatype cache
            term_value: The term value/text
            term_type: The term type ('uri', 'literal', 'blank')
            lang: Language tag for literals (optional)
            datatype_uri: Datatype URI for literals (optional)
            
        Returns:
            Tuple of (term_value, term_type, lang, datatype_id)
        """
        return await self.terms._process_term_with_datatype(space_id, term_value, term_type, lang, datatype_uri)

    def _initialize_space_datatype_cache_sync(self, space_id: str) -> None:
        """
        Initialize the datatype cache for a specific space by loading standard datatypes
        and any existing datatypes from the database (synchronous version).
        
        Args:
            space_id: Space identifier
        """
        return self.datatypes.initialize_space_datatype_cache_sync(space_id)

    async def add_rdf_quad(self, space_id: str, quad: Union[tuple, list]) -> bool:
        """
        Add an RDF quad to a specific space with proper datatype handling.
        
        This method uses the new datatype cache system to properly resolve and store
        datatype IDs for all literal terms in the quad.
        
        Args:
            space_id: Space identifier
            quad: Tuple of (subject, predicate, object, graph) RDF values
            
        Returns:
            bool: True if successful, False otherwise
        """
        return await self.db_ops.add_rdf_quad(space_id, quad)
    
    async def remove_rdf_quad(self, space_id: str, s: str, p: str, o: str, g: str) -> bool:
        """
        Remove an RDF quad from a specific space with proper datatype handling.
        
        This method uses the new datatype cache system to properly resolve datatypes
        when looking up terms for removal.
        
        Args:
            space_id: Space identifier
            s: Subject value (URI, literal, or blank node)
            p: Predicate value (typically URI)
            o: Object value (URI, literal, or blank node)
            g: Graph value (URI, literal, or blank node)
            
        Returns:
            bool: True if successful, False otherwise
        """
        return await self.db_ops.remove_rdf_quad(space_id, s, p, o, g)
    
    async def create_space_tables(self, space_id: str) -> bool:
        """
        Create UUID-based RDF tables optimized for deterministic batch loading.
        
        Args:
            space_id: Space identifier
            
        Returns:
            bool: True if tables created successfully, False otherwise
        """
        PostgreSQLSpaceUtils.validate_space_id(space_id)
        
        # Delegate table creation to database management class
        success = await self.db_mgmt.create_space_tables(space_id)
        
        if success:
            # Insert standard datatypes after table creation
            if not await self._insert_standard_datatypes(space_id):
                self.logger.warning(f"⚠️ Failed to insert standard datatypes for '{space_id}', but tables were created")
            
            # Initialize datatype cache for this space (async)
            await self._initialize_space_datatype_cache_sync(space_id)
        
        return success
    
    # SpaceBackendInterface adapter methods
    async def create_space_storage(self, space_id: str) -> bool:
        """Create space storage - adapter for SpaceBackendInterface."""
        return await self.create_space_tables(space_id)
    
    async def delete_space_storage(self, space_id: str) -> bool:
        """Delete space storage - adapter for SpaceBackendInterface."""
        return await self.delete_space_tables(space_id)
    
    async def drop_indexes_for_bulk_load(self, space_id: str) -> bool:
        """
        Drop all indexes before bulk loading for maximum performance.
        Keeps only primary keys and constraints.
        
        Args:
            space_id: Space identifier
            
        Returns:
            bool: True if indexes dropped successfully, False otherwise
        """
        # Delegate to database management class
        return await self.db_mgmt.drop_indexes_for_bulk_load(space_id)
    
    async def recreate_indexes_after_bulk_load(self, space_id: str, concurrent: bool = True) -> bool:
        """
        Recreate all indexes after bulk loading is complete.
        
        Args:
            space_id: Space identifier
            concurrent: If True, use CONCURRENTLY to avoid blocking queries
            
        Returns:
            bool: True if indexes recreated successfully, False otherwise
        """
        # Delegate to database management class
        return await self.db_mgmt.recreate_indexes_after_bulk_load(space_id, concurrent)
    
    async def delete_space_tables(self, space_id: str) -> bool:
        """
        Delete all RDF tables for a specific space.
        
        Args:
            space_id: Space identifier
            
        Returns:
            bool: True if tables deleted successfully, False otherwise
        """
        PostgreSQLSpaceUtils.validate_space_id(space_id)
        
        # Delegate to database management class
        return await self.db_mgmt.delete_space_tables(space_id)
    
    async def space_exists(self, space_id: str) -> bool:
        """
        Check if tables for a space exist in the database.
        
        Args:
            space_id: Space identifier
            
        Returns:
            bool: True if space tables exist, False otherwise
        """
        return await self.core.space_exists(space_id, self)
    
    async def list_spaces(self) -> List[str]:
        """
        List all spaces that have tables in the database.
        
        Returns:
            list: List of space IDs
        """
        return await self.core.list_spaces()
    
    async def get_space_info(self, space_id: str) -> Dict[str, Any]:
        """
        Get information about a specific space.
        
        Args:
            space_id: Space identifier
            
        Returns:
            dict: Space information including table details
        """
        return await self.core.get_space_info(space_id, self)
    
    async def add_term(self, space_id: str, term_text: str, term_type: str, 
                      lang: Optional[str] = None, datatype_id: Optional[int] = None) -> Optional[str]:
        """
        Add a term to the terms dictionary for a specific space using UUID-based approach.
        
        Args:
            space_id: Space identifier
            term_text: The term text (URI, literal value, etc.)
            term_type: Term type ('U'=URI, 'L'=Literal, 'B'=BlankNode, 'G'=Graph)
            lang: Language tag for literals
            datatype_id: Reference to datatype term ID
            
        Returns:
            str: Term UUID if successful, None otherwise
        """
        return await self.terms.add_term(space_id, term_text, term_type, lang, datatype_id)
    
    async def get_term_uuid(self, space_id: str, term_text: str, term_type: str, 
                           lang: Optional[str] = None, datatype_id: Optional[int] = None) -> Optional[str]:
        """
        Get term UUID for existing term in a specific space using datatype-aware approach.
        
        This method is maintained for backward compatibility but now uses the new
        datatype-aware infrastructure internally.
        
        Args:
            space_id: Space identifier
            term_text: The term text
            term_type: Term type
            lang: Language tag
            datatype_id: Datatype term ID (for backward compatibility)
            
        Returns:
            str: Term UUID if found, None otherwise
        """
        return await self.terms.get_term_uuid(space_id, term_text, term_type, lang, datatype_id)
    
    async def get_term_uuid_from_rdf_value(self, space_id: str, rdf_value) -> Optional[str]:
        """
        Get term UUID for existing term from an RDF value using datatype-aware approach.
        
        This method uses the new datatype cache system to properly resolve and match
        datatype IDs for literal terms.
        
        Args:
            space_id: Space identifier
            rdf_value: RDF value (URI, literal, or blank node)
            
        Returns:
            str: Term UUID if found, None otherwise
        """
        return await self.terms.get_term_uuid_from_rdf_value(space_id, rdf_value)
    
    async def delete_term(self, space_id: str, term_text: str, term_type: str, 
                         lang: Optional[str] = None, datatype_id: Optional[int] = None) -> bool:
        """
        Delete a term from a specific space using UUID-based approach.
        
        Note: This will only delete the term if it's not referenced by any quads.
        Use with caution as this could break referential integrity if not checked properly.
        
        Args:
            space_id: Space identifier
            term_text: The term text
            term_type: Term type
            lang: Language tag
            datatype_id: Datatype term ID
            
        Returns:
            bool: True if term was deleted, False otherwise
        """
        return await self.terms.delete_term(space_id, term_text, term_type, lang, datatype_id)
    
    # Quad management methods
    async def add_quad(self, space_id: str, subject_uuid: str, predicate_uuid: str, 
                      object_uuid: str, context_uuid: str) -> bool:
        """
        Add an RDF quad to a specific space using UUID-based approach.
        
        Args:
            space_id: Space identifier
            subject_uuid: Subject term UUID
            predicate_uuid: Predicate term UUID
            object_uuid: Object term UUID
            context_uuid: Context (graph) term UUID
            
        Returns:
            bool: True if successful, False otherwise
        """
        return await self.db_ops.add_quad(space_id, subject_uuid, predicate_uuid, object_uuid, context_uuid)
    
    async def get_quad_count(self, space_id: str, context_uuid: Optional[str] = None) -> int:
        """
        Get count of quads in a specific space using UUID-based approach, optionally filtered by context.
        
        Args:
            space_id: Space identifier
            context_uuid: Optional context UUID to filter by
            
        Returns:
            int: Number of quads
            
        Raises:
            Exception: If the space does not exist
        """
        return await self.queries.get_quad_count(space_id, context_uuid)
    
    async def get_rdf_quad_count(self, space_id: str, graph_uri: Optional[str] = None) -> int:
        """
        Get count of RDF quads in a specific space, optionally filtered by graph URI (context).
        
        This is a high-level RDF API that accepts graph URIs and converts them to UUIDs
        internally for compatibility with the UUID-based get_quad_count method.
        
        Args:
            space_id: Space identifier
            graph_uri: Optional graph URI to filter by (e.g., 'http://vital.ai/graph/test')
            
        Returns:
            int: Number of quads
            
        Raises:
            Exception: If the space does not exist
        """
        return await self.queries.get_rdf_quad_count(space_id, graph_uri)
    
    async def remove_quad(self, space_id: str, subject_uuid: str, predicate_uuid: str, object_uuid: str, context_uuid: str) -> bool:
        """
        Remove a single RDF quad from a specific space using UUID-based approach.
        
        Following RDFLib pattern: removes only one instance of the matching quad,
        not all instances. If multiple identical quads exist, only one is removed.
        
        Args:
            space_id: Space identifier
            subject_uuid: Subject term UUID
            predicate_uuid: Predicate term UUID
            object_uuid: Object term UUID
            context_uuid: Context (graph) term UUID
            
        Returns:
            bool: True if a quad was removed, False if no matching quad found
        """
        return await self.db_ops.remove_quad(space_id, subject_uuid, predicate_uuid, object_uuid, context_uuid)
    

    async def get_rdf_quad(self, space_id: str, s: str, p: str, o: str, g: str) -> bool:
        """
        Check if an RDF quad exists in a specific space using datatype-aware approach.
        
        This method uses the new datatype cache system to properly resolve and match
        datatype IDs for all literal terms in the quad.
        
        Args:
            space_id: Space identifier
            s: Subject value (URI, literal, or blank node)
            p: Predicate value (typically URI)
            o: Object value (URI, literal, or blank node)
            g: Graph/context value (typically URI)
            
        Returns:
            bool: True if the quad exists, False otherwise
        """
        return await self.queries.get_rdf_quad(space_id, s, p, o, g)
    
    async def add_rdf_quads_batch(self, space_id: str, quads: List[Tuple[Identifier, Identifier, Identifier, Identifier]], 
                                 auto_commit: bool = True, verify_count: bool = False, connection=None) -> int:
        """
        Datatype-aware batch RDF quad insertion with proper datatype handling.
        
        This method uses the new datatype cache system to properly resolve and store
        datatype IDs for all literal terms in the batch.
        
        Args:
            space_id: The space identifier
            quads: List of (subject, predicate, object, context) tuples
            auto_commit: Whether to commit the transaction automatically (default: True)
            verify_count: Whether to verify insertion with COUNT query (default: False, for performance)
            connection: Optional connection to use (for transaction management)
            
        Returns:
            Number of quads successfully inserted
        """
        return await self.db_ops.add_rdf_quads_batch(space_id, quads, auto_commit, verify_count, connection)
       
    async def remove_rdf_quads_batch(self, space_id: str, quads: List[tuple]) -> int:
        """
        Datatype-aware batch RDF quad removal with proper datatype handling.
        
        This method uses the new datatype cache system to properly resolve and match
        datatype IDs for all literal terms in the batch.
        
        Args:
            space_id: Space identifier
            quads: List of (s, p, o, g) tuples representing RDF quads to remove
            
        Returns:
            int: Number of quads successfully removed
        """
        return await self.db_ops.remove_rdf_quads_batch(space_id, quads)
    
    # Namespace management methods
    async def add_namespace(self, space_id: str, prefix: str, namespace_uri: str) -> Optional[int]:
        """
        Add a namespace prefix mapping to a specific space.
        
        Args:
            space_id: Space identifier
            prefix: Namespace prefix (e.g., 'foaf', 'rdf')
            namespace_uri: Full namespace URI (e.g., 'http://xmlns.com/foaf/0.1/')
            
        Returns:
            int: Namespace ID if successful, None otherwise
        """
        return await self.namespaces.add_namespace(space_id, prefix, namespace_uri)
    
    async def get_namespace_uri(self, space_id: str, prefix: str) -> Optional[str]:
        """
        Get namespace URI for a given prefix in a specific space.
        
        Args:
            space_id: Space identifier
            prefix: Namespace prefix to look up
            
        Returns:
            str: Namespace URI if found, None otherwise
        """
        return await self.namespaces.get_namespace_uri(space_id, prefix)
    
    async def list_namespaces(self, space_id: str) -> List[Dict[str, Any]]:
        """
        Get all namespace prefix mappings for a specific space.
        
        Args:
            space_id: Space identifier
            
        Returns:
            list: List of namespace dictionaries with id, prefix, namespace_uri, created_time
        """
        return await self.namespaces.list_namespaces(space_id)
    
    async def quads(self, space_id: str, quad_pattern: tuple, context: Optional[Any] = None):
        """
        A generator over all the quads matching the pattern. Pattern can include any objects
        for comparing against nodes in the store, including URIRef, Literal, BNode, Variable, 
        and REGEXTerm for regex pattern matching.
        
        This follows the RDFLib triples() pattern but for quads (subject, predicate, object, context).
        Used by SPARQL query implementation for quad pattern matching.
        
        REGEXTerm Support:
        - REGEXTerm instances in any position enable PostgreSQL regex matching (~)
        - Uses PostgreSQL's built-in regex engine for efficient pattern matching
        - Leverages pg_trgm indexes for optimized regex query performance
        - Supports full POSIX regex syntax in pattern strings
        
        Args:
            space_id: Space identifier
            quad_pattern: 4-tuple of (subject, predicate, object, context) patterns
            context: Optional context (not used in current implementation)
            
        Yields:
            tuple: (quad, context_iterator) where quad is (s, p, o, c) and 
                   context_iterator is a function that yields the context
        """
        # Delegate to the queries class
        async for result in self.queries.quads(space_id, quad_pattern, context):
            yield result
    
    def get_pool_stats(self) -> dict:
        """
        Get connection pool statistics for shared or dedicated RDF psycopg3 ConnectionPool.
        
        Returns:
            dict: Pool statistics including size, available connections, etc.
        """
        return self.core.get_pool_stats()
    
    def close_pool(self) -> None:
        """
        Close the connection pool and all connections.
        """
        self.core.close_pool()
        
     
    



import os
import json
import logging
from datetime import datetime, timedelta
from typing import Dict, List, Optional, Any
from pathlib import Path
from fastapi import FastAPI, Request, Response, HTTPException, status, Depends, Form, Body, WebSocket
from fastapi.responses import HTMLResponse, RedirectResponse, JSONResponse, FileResponse
from fastapi.staticfiles import StaticFiles
from fastapi.middleware.cors import CORSMiddleware
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from starlette.middleware.sessions import SessionMiddleware
from pydantic import BaseModel, Field
import uvicorn

# Import VitalGraph configuration loader
from vitalgraph.config.config_loader import get_config, ConfigurationError
from vitalgraph.api.vitalgraph_api import VitalGraphAPI
from vitalgraph.auth.vitalgraph_auth import VitalGraphAuth
from vitalgraph.impl.vitalgraph_impl import VitalGraphImpl
from vitalgraph.websocket.websocket_handler import ConnectionManager, websocket_endpoint
from vitalgraph.signal.notification_bridge import setup_notification_bridge
from vitalgraph.model.users_model import User, UsersListResponse, UserCreateResponse, UserUpdateResponse, UserDeleteResponse
from vitalgraph.model.spaces_model import Space, SpacesListResponse, SpaceCreateResponse, SpaceUpdateResponse, SpaceDeleteResponse


# Space and User models now imported from their respective model files


class VitalGraphAppImpl:
    """
    VitalGraph FastAPI Application Implementation
    
    This class implements the FastAPI application for VitalGraph, providing:
    - REST API endpoints for graph operations
    - WebSocket support for real-time updates
    - Frontend serving in production mode
    """
    
    def __init__(self, app: FastAPI, config: Dict):
        """Initialize VitalGraph application with FastAPI app and configuration."""
        self.app = app
        self.config = config
        
        # Configure logging based on config file
        try:
            app_config = config.get_app_config()
            log_level = app_config.get('log_level', 'INFO')
            numeric_level = getattr(logging, log_level.upper(), logging.INFO)
            
            # Set level for root logger
            logging.getLogger().setLevel(numeric_level)
            
            # Set level for all existing loggers
            for logger_name in logging.Logger.manager.loggerDict:
                logger = logging.getLogger(logger_name)
                logger.setLevel(numeric_level)
        except (AttributeError, TypeError):
            # Fallback to INFO if config access fails
            numeric_level = logging.INFO
            logging.getLogger().setLevel(logging.INFO)
        
        self.logger = logging.getLogger(f"{__name__}.VitalGraphAppImpl")
        self.logger.setLevel(numeric_level)  # Set this logger's level explicitly
        self.app_mode = os.getenv("APP_MODE", "development").lower()
        
        # Initialize VitalGraphImpl which handles configuration and database initialization
        self.vital_graph_impl = VitalGraphImpl(config=self.config)
        # Get database implementation from VitalGraphImpl
        self.db_impl = self.vital_graph_impl.get_db_impl()
        
        # Update config reference in case it was loaded by VitalGraphImpl
        self.config = self.vital_graph_impl.get_config()
        
        # Initialize authentication
        # Get JWT secret key from environment variable (required)
        jwt_secret = os.getenv("JWT_SECRET_KEY")
        if not jwt_secret:
            raise ValueError("JWT_SECRET_KEY environment variable is required but not set")
        
        self.logger.info("✅ Using JWT secret key from environment variable")
        self.auth = VitalGraphAuth(secret_key=jwt_secret)
        
        # Initialize WebSocket connection manager
        self.websocket_manager = ConnectionManager(self.auth)
        
        # Get Space Manager from VitalGraphImpl
        self.space_manager = self.vital_graph_impl.get_space_manager()
        print(f"🔍 DEBUG: Retrieved space_manager from VitalGraphImpl: {self.space_manager}")
        print(f"🔍 DEBUG: space_manager type: {type(self.space_manager)}")
        
        # Initialize VitalGraph API with auth handler, database implementation, and space manager
        self.api = VitalGraphAPI(
            auth_handler=self.auth,
            db_impl=self.db_impl, 
            space_manager=self.space_manager
        )
        print(f"🔍 DEBUG: VitalGraphAPI initialized with space_manager: {self.api.space_manager}")
        
        # Add dependency for getting current user (needed before route setup)
        self.get_current_user = self.auth.create_get_current_user_dependency()
        
        # Setup middleware, static files, startup events, and routes
        self._setup_middleware()
        self._setup_static_files()
        self._setup_startup_events()
        
        # Setup endpoints with API instance (space_manager will be set later in startup event)
        self._setup_all_endpoints()

    def _setup_middleware(self):
        """Setup CORS and session middleware"""
        
        # Add request logging middleware
        @self.app.middleware("http")
        async def log_requests(request: Request, call_next):
            self.logger.debug(f"🔍 INCOMING REQUEST: {request.method} {request.url}")
            self.logger.debug(f"🔍 REQUEST HEADERS: {dict(request.headers)}")
            response = await call_next(request)
            self.logger.debug(f"🔍 RESPONSE STATUS: {response.status_code}")
            return response
        
        # Add CORS middleware
        self.app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],  # In production, specify exact origins
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
        # Add session middleware for authentication
        self.app.add_middleware(
            SessionMiddleware,
            secret_key=os.getenv("SESSION_SECRET_KEY", "your-secret-key-here"),
            max_age=86400  # 24 hours
        )

    def _setup_static_files(self):
        """Mount React frontend build files only in production"""
        if self.app_mode == "production":
            # Mount static files from React build
            project_root = Path(__file__).parent.parent.parent
            static_dir = project_root / "vitalgraph" / "api" / "frontend" / "dist"
            
            if static_dir.exists():
                self.app.mount("/static", StaticFiles(directory=str(static_dir)), name="static")
            else:
                self.logger.warning(f"Static files directory not found at {static_dir}")

    def _setup_startup_events(self):
        """Setup FastAPI startup events"""
        @self.app.on_event("startup")
        async def startup_event():
            """Connect to database on server startup"""
            if self.vital_graph_impl:
                try:
                    self.logger.debug("About to connect database via VitalGraphImpl")
                    # Use VitalGraphImpl.connect_database() which handles SignalManager creation and injection
                    success = await self.vital_graph_impl.connect_database()
                    self.logger.info(f"Connected to database successfully: {success}")
                    
                    # Update space_manager reference after database connection
                    self.space_manager = self.vital_graph_impl.get_space_manager()
                    print(f"🔍 STARTUP DEBUG: Updated space_manager after DB connection: {self.space_manager}")
                    
                    # Set the space_manager on VitalGraphAPI now that it's available
                    if self.space_manager is not None:
                        self.api.space_manager = self.space_manager
                        print(f"🔍 STARTUP DEBUG: Set space_manager on VitalGraphAPI: {self.api.space_manager}")
                        print(f"🔍 STARTUP DEBUG: API object id: {id(self.api)}")
                        print(f"🔍 STARTUP DEBUG: Endpoints already registered during init - they will use this updated API instance")
                    else:
                        print(f"❌ STARTUP ERROR: space_manager is still None after database connection")
                    
                    # Ensure SpaceManager is initialized from database after connection
                    self.logger.debug("Checking SpaceManager for initialization...")
                    self.logger.debug(f"self.space_manager exists: {self.space_manager is not None}")
                    self.logger.debug(f"SpaceManager class: {type(self.space_manager)}")
                    self.logger.debug(f"SpaceManager module: {type(self.space_manager).__module__}")
                    if self.space_manager is not None:
                        self.logger.debug(f"Available methods: {[m for m in dir(self.space_manager) if not m.startswith('_')]}")
                        self.logger.debug(f"has initialize_from_database: {hasattr(self.space_manager, 'initialize_from_database')}")
                    
                    if self.space_manager is not None and hasattr(self.space_manager, 'initialize_from_database'):
                        self.logger.debug("Calling initialize_from_database()...")
                        await self.space_manager.initialize_from_database()
                        self.logger.info(f"SpaceManager initialized with {len(self.space_manager)} spaces")
                        self.logger.debug(f"Available spaces: {list(self.space_manager._spaces.keys()) if hasattr(self.space_manager, '_spaces') else 'N/A'}")
                    else:
                        self.logger.warning("SpaceManager initialization skipped - conditions not met")
                    
                    # Endpoints are already initialized in __init__
                    # SpaceManager is now fully initialized and connected
                    self.logger.info("SpaceManager ready for endpoint operations")
                    
                    # Set up notification bridge between PostgreSQL and WebSocket
                    if self.websocket_manager is not None:
                        try:
                            # Debug prints to investigate references
                            self.logger.debug(f"VitalGraphAppImpl db_impl id: {id(self.db_impl)}")
                            self.logger.debug(f"VitalGraphImpl db_impl id: {id(self.vital_graph_impl.db_impl)}")
                            
                            # Try to access signal_manager from VitalGraphImpl directly
                            signal_manager = self.vital_graph_impl.signal_manager
                            self.logger.debug(f"Using VitalGraphImpl signal_manager: {signal_manager}")
                            
                            if signal_manager is None:
                                # Fallback to db_impl.get_signal_manager() if direct access fails
                                signal_manager = self.db_impl.get_signal_manager()
                                self.logger.debug(f"Using db_impl.get_signal_manager(): {signal_manager}")
                                
                            self.logger.debug("Setting up notification bridge between PostgreSQL and WebSocket...")
                            bridge = await setup_notification_bridge(signal_manager, self.websocket_manager)
                            if bridge:
                                self.logger.info("PostgreSQL notification bridge initialized successfully")
                            else:
                                self.logger.warning("Failed to initialize notification bridge")
                        except Exception as e:
                            self.logger.warning(f"Error setting up notification bridge: {e}")
                    else:
                        self.logger.warning("WebSocket manager not available, notification bridge skipped")
                    
                except Exception as e:
                    self.logger.error(f"Failed to connect to database: {e}")
    
    def _setup_all_endpoints(self):
        """Setup all API endpoints in a single atomic operation."""
        self.logger.info("Setting up all endpoints...")
        
        # Guard against duplicate initialization
        if hasattr(self, '_all_endpoints_initialized'):
            self.logger.info("Endpoints already initialized, skipping...")
            return
        self._all_endpoints_initialized = True
        
        # Initialize all endpoint groups in order (most specific routes first)
        self.logger.info("Initializing auth routes...")
        self._init_auth_routes()
        self.logger.info("Initializing space routes...")
        self._init_space_routes()
        self.logger.info("Initializing user routes...")
        self._init_user_routes()
        self.logger.info("Initializing websocket routes...")
        self._init_websocket_routes()
        self.logger.info("Initializing SPARQL routers...")
        self._init_sparql_routers()
        self.logger.info("Initializing graph data routers...")
        self._init_graph_data_routers()
        self.logger.info("Initializing data routers...")
        self._init_data_routers()
        self.logger.info("Initializing frontend routes...")
        self._init_frontend_routes()
        self.logger.info("All endpoints initialized successfully!")
    
    def _init_sparql_routers(self):
        """Initialize SPARQL endpoint routers."""
        # Set logger to DEBUG level for this function
        import logging
        self.logger.setLevel(logging.DEBUG)
        
        # Import SPARQL endpoint routers
        from vitalgraph.endpoint.sparql_query_endpoint import create_sparql_query_router
        from vitalgraph.endpoint.sparql_update_endpoint import create_sparql_update_router
        from vitalgraph.endpoint.sparql_insert_endpoint import create_sparql_insert_router
        from vitalgraph.endpoint.sparql_delete_endpoint import create_sparql_delete_router
        from vitalgraph.endpoint.sparql_graph_endpoint import create_sparql_graph_router
        
        # Create routers with space manager and auth dependency
        self.logger.info("Creating SPARQL routers...")
        query_router = create_sparql_query_router(self.space_manager, self.get_current_user)
        update_router = create_sparql_update_router(self.space_manager, self.get_current_user)
        insert_router = create_sparql_insert_router(self.space_manager, self.get_current_user)
        delete_router = create_sparql_delete_router(self.space_manager, self.get_current_user)
        graph_router = create_sparql_graph_router(self.space_manager, self.get_current_user)
        
        # Include routers in the FastAPI app
        self.logger.info("Including SPARQL routers in FastAPI app...")
        self.app.include_router(query_router, prefix="/api/graphs/sparql", tags=["SPARQL"])
        self.app.include_router(update_router, prefix="/api/graphs/sparql", tags=["SPARQL"])
        self.app.include_router(insert_router, prefix="/api/graphs/sparql", tags=["SPARQL"])
        self.app.include_router(delete_router, prefix="/api/graphs/sparql")
        self.app.include_router(graph_router, prefix="/api/graphs/sparql")
        self.logger.info(f"Included graph router with prefix '/api/graphs/sparql' - {len(graph_router.routes)} routes")
    
    def _init_graph_data_routers(self):
        """Initialize graph data endpoint routers."""
        # Import graph data endpoint routers
        from vitalgraph.endpoint.triples_endpoint import create_triples_router
        from vitalgraph.endpoint.kgtypes_endpoint import create_kgtypes_router
        from vitalgraph.endpoint.objects_endpoint import create_objects_router
        from vitalgraph.endpoint.kgentities_endpoint import create_kgentities_router
        from vitalgraph.endpoint.kgframes_endpoint import create_kgframes_router
        from vitalgraph.endpoint.kgrelations_endpoint import create_kgrelations_router
        from vitalgraph.endpoint.kgquery_endpoint import create_kgqueries_router
        from vitalgraph.endpoint.files_endpoint import create_files_router
        
        # Create routers with space manager and auth dependency
        triples_router = create_triples_router(self.space_manager, self.get_current_user)
        kgtypes_router = create_kgtypes_router(self.space_manager, self.get_current_user)
        objects_router = create_objects_router(self.space_manager, self.get_current_user)
        kgentities_router = create_kgentities_router(self.space_manager, self.get_current_user)
        kgframes_router = create_kgframes_router(self.space_manager, self.get_current_user)
        kgrelations_router = create_kgrelations_router(self.space_manager, self.get_current_user)
        kgqueries_router = create_kgqueries_router(self.space_manager, self.get_current_user)
        files_router = create_files_router(self.space_manager, self.get_current_user, config=self.config)
        
        # Include routers in the FastAPI app  
        self.app.include_router(triples_router, prefix="/api/graphs")
        self.app.include_router(kgtypes_router, prefix="/api/graphs")
        self.app.include_router(objects_router, prefix="/api/graphs")
        self.app.include_router(kgentities_router, prefix="/api/graphs")
        self.app.include_router(kgframes_router, prefix="/api/graphs")
        self.app.include_router(kgrelations_router, prefix="/api/graphs")
        self.app.include_router(kgqueries_router, prefix="/api/graphs")
        self.app.include_router(files_router, prefix="/api")
    
    def _init_data_routers(self):
        """Initialize data import/export endpoint routers."""
        # Import data endpoint routers
        from vitalgraph.endpoint.import_endpoint import create_import_router
        from vitalgraph.endpoint.export_endpoint import create_export_router
        
        # Create routers with space manager and auth dependency
        import_router = create_import_router(self.space_manager, self.get_current_user)
        export_router = create_export_router(self.space_manager, self.get_current_user)
        
        # Include routers in the FastAPI app
        self.app.include_router(import_router, prefix="/api/data", tags=["Data"])
        self.app.include_router(export_router, prefix="/api/data", tags=["Data"])
    
    def _init_auth_routes(self):
        """Initialize authentication routes."""
        # Authentication wrapper functions
        async def logout_wrapper(request: Request, current_user: Dict = Depends(self.get_current_user)):
            return await self.logout(request, current_user)
        
        async def refresh_token_wrapper(refresh_token: str = Body(..., embed=True), current_user: Dict = Depends(self.get_current_user)):
            return await self.refresh_token(refresh_token)

        # Authentication routes
        self.app.post(
            "/api/login",
            tags=["Authentication"],
            summary="User Login",
            description="Authenticate user and generate access token for API access"
        )(self.login)
        self.app.post(
            "/api/logout",
            tags=["Authentication"],
            summary="User Logout",
            description="Invalidate current session and log out user"
        )(logout_wrapper)
        self.app.post(
            "/api/refresh",
            tags=["Authentication"],
            summary="Refresh Token",
            description="Refresh access token using refresh token (requires refresh token as Bearer token)"
        )(refresh_token_wrapper)
    
    def _init_space_routes(self):
        """Initialize space management routes."""
        from vitalgraph.endpoint.spaces_endpoint import create_spaces_router
        
        # Create and include spaces router
        spaces_router = create_spaces_router(self.api, self.get_current_user)
        self.app.include_router(spaces_router, prefix="/api")
    
    def _init_user_routes(self):
        """Initialize user management routes."""
        from vitalgraph.endpoint.users_endpoint import create_users_router
        
        # Create and include users router
        users_router = create_users_router(self.api, self.get_current_user)
        self.app.include_router(users_router, prefix="/api")
    
    def _init_websocket_routes(self):
        """Initialize WebSocket routes."""
        async def websocket_wrapper(websocket: WebSocket):
            await websocket_endpoint(websocket, self.websocket_manager)

        self.app.websocket("/api/ws")(websocket_wrapper)
    
    def _init_frontend_routes(self):
        """Initialize frontend serving routes."""
        # Frontend serving routes
        if self.app_mode == "production":
            self.app.get("/", response_class=HTMLResponse)(self.serve_frontend)
            self.app.get("/{path:path}")(self.catch_all)
        else:
            self.app.get("/")(self.api_root)
    

    async def login(self, form_data: OAuth2PasswordRequestForm = Depends(), token_expiry_seconds: Optional[int] = Form(None)):
        """Login endpoint - delegates to API class"""
        return await self.api.login(form_data, token_expiry_seconds)

    async def logout(self, request: Request, current_user: Dict):
        """Logout endpoint - delegates to API class"""
        return await self.api.logout(request, current_user)

    async def refresh_token(self, refresh_token: str = Body(..., embed=True)):
        """Refresh token endpoint - delegates to API class"""
        return await self.api.refresh_token(refresh_token)

    async def list_spaces(self, current_user: Dict):
        """Get list of spaces for the authenticated user"""
        return await self.api.list_spaces(current_user)

    async def sparql_endpoint(
        self,
        query: str,
        current_user: Dict
    ):
        """Execute a SPARQL query"""
        return await self.api.execute_sparql_query(query, current_user)

    async def health(self):
        """Health check endpoint - delegates to API class"""
        return await self.api.health()
    
    async def serve_frontend(self):
        """Serve the built React app's index.html"""
        project_root = Path(__file__).parent.parent.parent
        index_path = project_root / "vitalgraph" / "api" / "frontend" / "dist" / "index.html"
        if index_path.exists():
            with open(index_path) as f:
                return HTMLResponse(content=f.read())
        # Fallback to API docs if index.html doesn't exist
        return RedirectResponse(url="/docs")
    
    async def api_root(self):
        """Development mode root endpoint"""
        return {
            "message": "VitalGraph API Server",
            "documentation": "/docs",
            "alternative_docs": "/redoc"
        }
    
    async def catch_all(self, path: str):
        """Catch-all route for client-side routing - only in production"""
        project_root = Path(__file__).parent.parent.parent
        
        # First, try to serve static files from the dist directory (SVGs, favicon, etc.)
        static_file_path = project_root / "vitalgraph" / "api" / "frontend" / "dist" / path
        if static_file_path.exists() and static_file_path.is_file():
            return FileResponse(static_file_path)
        
        # Only catch routes that don't start with /api or /static for SPA routing
        if path.startswith("api/") or path.startswith("static/") or path.startswith("docs") or path == "openapi.json":
            raise HTTPException(status_code=404, detail="Not found")
        
        # Return the index.html for client-side routing
        index_path = project_root / "vitalgraph" / "api" / "frontend" / "dist" / "index.html"
        if index_path.exists():
            with open(index_path) as f:
                return HTMLResponse(content=f.read())
        return RedirectResponse(url="/docs")



#!/usr/bin/env python3
"""
KG SPARQL Utilities

Utility functions for SPARQL query processing and result handling.
Used across KG processors to provide consistent SPARQL operations.
"""

import logging
from typing import Dict, Any, List, Optional, Union
from dataclasses import dataclass

logger = logging.getLogger(__name__)


class KGSparqlUtils:
    """Utility class for common SPARQL operations and result processing."""
    
    @staticmethod
    def extract_count_from_results(results: Any) -> int:
        """
        Extract count value from SPARQL COUNT query results.
        
        Args:
            results: SPARQL query results
            
        Returns:
            int: Extracted count value, 0 if not found
        """
        try:
            if isinstance(results, dict) and results.get("results", {}).get("bindings"):
                bindings = results["results"]["bindings"]
                if bindings and len(bindings) > 0:
                    count_binding = bindings[0]
                    if "count" in count_binding:
                        count_value = count_binding["count"].get("value", "0")
                        return int(count_value)
            return 0
        except (ValueError, KeyError, TypeError) as e:
            logger.warning(f"Error extracting count from SPARQL results: {e}")
            return 0
    
    @staticmethod
    def extract_uris_from_results(results: Any, variable_name: str = "uri") -> List[str]:
        """
        Extract URI values from SPARQL query results.
        
        Args:
            results: SPARQL query results
            variable_name: Name of the variable containing URIs
            
        Returns:
            List[str]: List of extracted URIs
        """
        uris = []
        try:
            if isinstance(results, dict) and results.get("results", {}).get("bindings"):
                for binding in results["results"]["bindings"]:
                    if variable_name in binding:
                        uri_value = binding[variable_name].get("value")
                        if uri_value:
                            uris.append(uri_value)
        except (KeyError, TypeError) as e:
            logger.warning(f"Error extracting URIs from SPARQL results: {e}")
        
        return uris
    
    @staticmethod
    def extract_subject_uris_from_results(results: Any) -> List[str]:
        """
        Extract subject URIs from SPARQL query results.
        
        Args:
            results: SPARQL query results
            
        Returns:
            List[str]: List of extracted subject URIs
        """
        return KGSparqlUtils.extract_uris_from_results(results, "subject")
    
    @staticmethod
    def build_search_filter(search: Optional[str], search_properties: List[str] = None) -> str:
        """
        Build SPARQL search filter for text search across properties.
        
        Args:
            search: Search term
            search_properties: List of properties to search (defaults to common text properties)
            
        Returns:
            str: SPARQL filter clause
        """
        if not search:
            return ""
        
        if search_properties is None:
            search_properties = [
                "haley:hasName", 
                "vital:hasName", 
                "haley:hasKGSlotStringValue",
                "vital:hasDescription"
            ]
        
        # Escape search term for SPARQL
        escaped_search = search.replace('"', '\\"')
        
        # Build filter conditions for each property
        filter_conditions = []
        for prop in search_properties:
            filter_conditions.append(f'CONTAINS(LCASE(STR(?{prop.split(":")[-1]})), LCASE("{escaped_search}"))')
        
        if filter_conditions:
            return f"""
            OPTIONAL {{ ?subject {search_properties[0]} ?{search_properties[0].split(":")[-1]} }}
            """ + "".join([f"""
            OPTIONAL {{ ?subject {prop} ?{prop.split(":")[-1]} }}""" for prop in search_properties[1:]]) + f"""
            FILTER ({" || ".join(filter_conditions)})
            """
        
        return ""
    
    @staticmethod
    def build_pagination_clause(page_size: int, offset: int) -> str:
        """
        Build SPARQL LIMIT and OFFSET clause for pagination.
        
        Args:
            page_size: Number of results per page
            offset: Offset for pagination
            
        Returns:
            str: SPARQL pagination clause
        """
        return f"LIMIT {page_size} OFFSET {offset}"
    
    @staticmethod
    def build_graph_clause(graph_id: str) -> str:
        """
        Build SPARQL GRAPH clause.
        
        Args:
            graph_id: Graph identifier
            
        Returns:
            str: SPARQL GRAPH clause
        """
        return f"GRAPH <{graph_id}>"
    
    @staticmethod
    def escape_sparql_string(value: str) -> str:
        """
        Escape string value for use in SPARQL queries.
        
        Args:
            value: String value to escape
            
        Returns:
            str: Escaped string value
        """
        if not isinstance(value, str):
            return str(value)
        
        # Escape quotes and backslashes
        escaped = value.replace('\\', '\\\\').replace('"', '\\"')
        return escaped
    
    @staticmethod
    def build_uri_reference(uri: str) -> str:
        """
        Build proper URI reference for SPARQL queries.
        
        Args:
            uri: URI string
            
        Returns:
            str: Properly formatted URI reference
        """
        if not uri:
            return ""
        
        # If already wrapped in angle brackets, return as-is
        if uri.startswith('<') and uri.endswith('>'):
            return uri
        
        # Wrap in angle brackets
        return f"<{uri}>"
    
    @staticmethod
    def build_prefixes() -> str:
        """
        Build standard SPARQL prefixes for KG operations.
        
        Returns:
            str: SPARQL prefix declarations
        """
        return """
        PREFIX haley: <http://vital.ai/ontology/haley-ai-kg#>
        PREFIX vital: <http://vital.ai/ontology/vital-core#>
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        """
    
    @staticmethod
    def validate_sparql_results(results: Any) -> bool:
        """
        Validate SPARQL query results structure.
        
        Args:
            results: SPARQL query results
            
        Returns:
            bool: True if results are valid, False otherwise
        """
        try:
            if not isinstance(results, dict):
                return False
            
            if "results" not in results:
                return False
            
            if "bindings" not in results["results"]:
                return False
            
            return True
        except Exception:
            return False
    
    @staticmethod
    def build_type_filter(entity_type: Optional[str]) -> str:
        """
        Build SPARQL type filter clause.
        
        Args:
            entity_type: Entity type URI
            
        Returns:
            str: SPARQL type filter clause
        """
        if not entity_type:
            return ""
        
        return f"""
        ?subject a {KGSparqlUtils.build_uri_reference(entity_type)} .
        """
    
    @staticmethod
    def build_grouping_uri_filter(grouping_uri: str, property_name: str = "haley:kGGraphURI") -> str:
        """
        Build SPARQL filter for grouping URI properties.
        
        Args:
            grouping_uri: Grouping URI value
            property_name: Property name for grouping URI
            
        Returns:
            str: SPARQL grouping URI filter clause
        """
        if not grouping_uri:
            return ""
        
        return f"""
        ?subject {property_name} {KGSparqlUtils.build_uri_reference(grouping_uri)} .
        """
    
    @staticmethod
    def extract_triples_from_sparql_results(results: Any) -> List[tuple]:
        """
        Extract triples from SPARQL SELECT results.
        
        Args:
            results: SPARQL query results
            
        Returns:
            List[tuple]: List of (subject, predicate, object) triples
        """
        triples = []
        try:
            # Handle nested structure: results.results.bindings
            if isinstance(results, dict) and "results" in results and isinstance(results["results"], dict):
                bindings = results["results"].get("bindings", [])
                for binding in bindings:
                    subject = binding.get("subject", {}).get("value")
                    predicate = binding.get("predicate", {}).get("value")
                    obj = binding.get("object", {}).get("value")
                    if subject and predicate and obj:
                        triples.append((subject, predicate, obj))
            # Handle flat structure: results.bindings
            elif isinstance(results, dict) and results.get("bindings"):
                for binding in results["bindings"]:
                    subject = binding.get("subject", {}).get("value")
                    predicate = binding.get("predicate", {}).get("value")
                    obj = binding.get("object", {}).get("value")
                    if subject and predicate and obj:
                        triples.append((subject, predicate, obj))
            elif isinstance(results, list):
                for item in results:
                    if isinstance(item, dict):
                        subject = item.get("subject")
                        predicate = item.get("predicate")
                        obj = item.get("object")
                        
                        # Handle both string and dict formats
                        if isinstance(subject, dict):
                            subject = subject.get("value")
                        if isinstance(predicate, dict):
                            predicate = predicate.get("value")
                        if isinstance(obj, dict):
                            obj = obj.get("value")
                        
                        if subject and predicate and obj:
                            triples.append((subject, predicate, obj))
        except (KeyError, TypeError) as e:
            logger.warning(f"Error extracting triples from SPARQL results: {e}")
        
        return triples
    
    @staticmethod
    def extract_frame_uris_from_results(results: Any) -> List[str]:
        """
        Extract frame URIs from SPARQL query results.
        
        Args:
            results: SPARQL query results
            
        Returns:
            List[str]: List of extracted frame URIs
        """
        frame_uris = []
        try:
            if isinstance(results, dict) and results.get("bindings"):
                for binding in results["bindings"]:
                    frame_uri = binding.get("frame_uri", {}).get("value")
                    if frame_uri:
                        frame_uris.append(frame_uri)
            elif isinstance(results, list):
                for item in results:
                    if isinstance(item, dict) and "frame_uri" in item:
                        frame_value = item["frame_uri"]
                        if isinstance(frame_value, str):
                            frame_uris.append(frame_value)
                        elif isinstance(frame_value, dict):
                            frame_uri = frame_value.get("value")
                            if frame_uri:
                                frame_uris.append(frame_uri)
        except (KeyError, TypeError) as e:
            logger.warning(f"Error extracting frame URIs from SPARQL results: {e}")
        
        return frame_uris
    
    @staticmethod
    def convert_triples_to_vitalsigns_frames(triples: List[Dict[str, str]]) -> List:
        """
        Convert triples to VitalSigns frame objects using GraphObject.from_triples_list().
        
        Args:
            triples: List of triple dictionaries with 'subject', 'predicate', 'object' keys
            
        Returns:
            List: List of VitalSigns KGFrame objects
        """
        try:
            if not triples:
                return []
            
            # Convert triples to the format expected by GraphObject.from_triples_list()
            # The method expects a Generator[Tuple, None, None] of RDFLib (URIRef, URIRef, URIRef/Literal) tuples
            def triples_generator():
                from rdflib import URIRef, Literal
                for triple in triples:
                    subject = URIRef(triple["subject"])
                    predicate = URIRef(triple["predicate"])
                    
                    # Object might be a URI or literal value
                    obj_str = triple["object"]
                    from vital_ai_vitalsigns.utils.uri_utils import validate_rfc3986
                    if validate_rfc3986(obj_str, rule='URI'):
                        obj = URIRef(obj_str)
                    else:
                        obj = Literal(obj_str)
                    
                    yield (subject, predicate, obj)
            
            # Use GraphObject.from_triples_list() to convert all triples to VitalSigns objects
            from vital_ai_vitalsigns.model.GraphObject import GraphObject
            all_objects = GraphObject.from_triples_list(triples_generator())
            
            # Filter for KGFrame objects
            frames = []
            from ai_haley_kg_domain.model.KGFrame import KGFrame
            
            for obj in all_objects:
                if isinstance(obj, KGFrame):
                    frames.append(obj)
            
            logger.info(f"🔍 DEBUG: Converted {len(triples)} triples to {len(all_objects)} objects, {len(frames)} frames")
            
            return frames
            
        except Exception as e:
            logger.error(f"Error converting triples to VitalSigns frames: {e}")
            return []


class KGSparqlQueryBuilder:
    """Builder class for constructing common SPARQL queries."""
    
    def __init__(self):
        self.utils = KGSparqlUtils()
    
    def build_frame_discovery_query(self, graph_id: str, entity_uri: str, 
                                  page_size: int = 100, offset: int = 0, 
                                  search: Optional[str] = None) -> str:
        """
        Build SPARQL query to discover frames associated with an entity.
        
        Args:
            graph_id: Graph identifier
            entity_uri: Entity URI
            page_size: Number of results per page
            offset: Offset for pagination
            search: Optional search term
            
        Returns:
            str: SPARQL query string
        """
        search_filter = self.utils.build_search_filter(search) if search else ""
        pagination = self.utils.build_pagination_clause(page_size, offset)
        
        return f"""
        {self.utils.build_prefixes()}
        
        SELECT DISTINCT ?frame WHERE {{
            {self.utils.build_graph_clause(graph_id)} {{
                ?frame a haley:KGFrame ;
                       haley:hasKGGraphURI {self.utils.build_uri_reference(entity_uri)} .
                {search_filter}
            }}
        }}
        ORDER BY ?frame
        {pagination}
        """
    
    def build_frame_count_query(self, graph_id: str, entity_uri: str, 
                               search: Optional[str] = None) -> str:
        """
        Build SPARQL query to count frames associated with an entity.
        
        Args:
            graph_id: Graph identifier
            entity_uri: Entity URI
            search: Optional search term
            
        Returns:
            str: SPARQL count query string
        """
        search_filter = self.utils.build_search_filter(search) if search else ""
        
        return f"""
        {self.utils.build_prefixes()}
        
        SELECT (COUNT(DISTINCT ?frame) as ?count) WHERE {{
            {self.utils.build_graph_clause(graph_id)} {{
                ?frame a haley:KGFrame ;
                       haley:hasKGGraphURI {self.utils.build_uri_reference(entity_uri)} .
                {search_filter}
            }}
        }}
        """
    
    def build_frame_graph_query(self, graph_id: str, frame_uri: str, 
                               include_frame_graph: bool = True) -> str:
        """
        Build SPARQL query to get frame and optionally its complete graph.
        
        Args:
            graph_id: Graph identifier
            frame_uri: Frame URI
            include_frame_graph: Whether to include complete frame graph
            
        Returns:
            str: SPARQL query string
        """
        import logging
        logger = logging.getLogger(f"{__name__}.KGSparqlQueryBuilder")
        
        if include_frame_graph:
            # Get frame and all objects in its frame graph, including connecting edges
            query = f"""
            {self.utils.build_prefixes()}
            
            SELECT DISTINCT ?subject WHERE {{
                {self.utils.build_graph_clause(graph_id)} {{
                    {{
                        # Get the frame itself
                        BIND({self.utils.build_uri_reference(frame_uri)} as ?subject)
                        {self.utils.build_uri_reference(frame_uri)} a haley:KGFrame .
                    }} UNION {{
                        # Get all objects that belong to this frame's graph
                        ?subject haley:hasFrameGraphURI {self.utils.build_uri_reference(frame_uri)} .
                    }} UNION {{
                        # Get edges connecting objects within the frame
                        # Only include edges that have frameGraphURI (excludes Edge_hasEntityKGFrame)
                        ?subject haley:hasFrameGraphURI {self.utils.build_uri_reference(frame_uri)} .
                        ?subject vital:hasEdgeSource ?frameObject .
                    }}
                }}
            }}
            """
            logger.error(f"🔍 Built frame graph query for {frame_uri}")
            logger.error(f"🔍 Query will look for:")
            logger.error(f"🔍   1. Frame itself: {frame_uri}")
            logger.error(f"🔍   2. Objects with hasFrameGraphURI = {frame_uri}")
            logger.error(f"🔍   3. Edges with hasFrameGraphURI = {frame_uri}")
            return query
        else:
            # Get only the frame itself
            return f"""
            {self.utils.build_prefixes()}
            
            SELECT DISTINCT ?subject WHERE {{
                {self.utils.build_graph_clause(graph_id)} {{
                    BIND({self.utils.build_uri_reference(frame_uri)} as ?subject)
                    {self.utils.build_uri_reference(frame_uri)} a haley:KGFrame .
                }}
            }}
            """
    
    def build_frame_deletion_count_query(self, graph_id: str, frame_uri: str) -> str:
        """
        Build SPARQL query to count objects that will be deleted with a frame.
        
        Args:
            graph_id: Graph identifier
            frame_uri: Frame URI
            
        Returns:
            str: SPARQL count query string
        """
        return f"""
        {self.utils.build_prefixes()}
        
        SELECT (COUNT(?subject) as ?count) WHERE {{
            {self.utils.build_graph_clause(graph_id)} {{
                {{
                    # Count the frame itself
                    BIND({self.utils.build_uri_reference(frame_uri)} as ?subject)
                    {self.utils.build_uri_reference(frame_uri)} a haley:KGFrame .
                }} UNION {{
                    # Count all objects that belong to this frame's graph
                    ?subject haley:hasFrameGraphURI {self.utils.build_uri_reference(frame_uri)} .
                }}
            }}
        }}
        """
    
    def build_frame_deletion_query(self, graph_id: str, frame_uri: str) -> str:
        """
        Build SPARQL DELETE query to remove frame and its complete graph.
        
        Args:
            graph_id: Graph identifier
            frame_uri: Frame URI
            
        Returns:
            str: SPARQL DELETE query string
        """
        return f"""
        {self.utils.build_prefixes()}
        
        DELETE {{
            {self.utils.build_graph_clause(graph_id)} {{
                ?subject ?predicate ?object .
            }}
        }}
        WHERE {{
            {self.utils.build_graph_clause(graph_id)} {{
                {{
                    # Delete the frame itself
                    BIND({self.utils.build_uri_reference(frame_uri)} as ?subject)
                    {self.utils.build_uri_reference(frame_uri)} ?predicate ?object .
                }} UNION {{
                    # Delete all objects that belong to this frame's graph
                    ?subject haley:hasFrameGraphURI {self.utils.build_uri_reference(frame_uri)} ;
                             ?predicate ?object .
                }}
            }}
        }}
        """
    
    def build_entity_graphs_query(self, graph_id: str, entity_type_uri: Optional[str] = None, 
                                 search: Optional[str] = None, page_size: int = 100, 
                                 offset: int = 0) -> str:
        """
        Build SPARQL query for listing complete entity graphs (all objects with kGGraphURI).
        
        Args:
            graph_id: Graph identifier
            entity_type_uri: Optional entity type filter
            search: Optional search term
            page_size: Number of results per page
            offset: Offset for pagination
            
        Returns:
            str: SPARQL query string
        """
        # Build entity type filter
        if entity_type_uri:
            entity_filter = f"?entity a {self.utils.build_uri_reference(entity_type_uri)} ."
        else:
            entity_filter = """
            ?entity a ?entityType .
            FILTER(STRSTARTS(STR(?entityType), "http://vital.ai/ontology/haley-ai-kg#KG") && STRENDS(STR(?entityType), "Entity"))
            """
        
        # Build search filter
        search_filter = ""
        if search:
            escaped_search = self.utils.escape_sparql_string(search)
            search_filter = f"""
            ?entity vital:hasName ?name .
            FILTER(CONTAINS(LCASE(?name), LCASE("{escaped_search}")))
            """
        
        # Build pagination
        pagination = self.utils.build_pagination_clause(page_size, offset)
        
        return f"""
        {self.utils.build_prefixes()}
        
        SELECT DISTINCT ?subject WHERE {{
            {self.utils.build_graph_clause(graph_id)} {{
                # Find entities first
                {entity_filter}
                {search_filter}
                
                # Then find all objects with matching hasKGGraphURI
                ?subject haley:hasKGGraphURI ?entity .
            }}
        }}
        ORDER BY ?subject
        {pagination}
        """
    
    def build_list_entities_query(self, graph_id: str, entity_type_uri: Optional[str] = None, 
                                 search: Optional[str] = None, page_size: int = 100, 
                                 offset: int = 0) -> str:
        """
        Build SPARQL query for listing entity subjects (subject-first pattern).
        
        Args:
            graph_id: Graph identifier
            entity_type_uri: Optional entity type filter
            search: Optional search term
            page_size: Number of results per page
            offset: Offset for pagination
            
        Returns:
            str: SPARQL query string
        """
        # Build entity type filter
        # KGEntity uses hasKGEntityType property, not rdf:type
        if entity_type_uri:
            type_filter = f"""
            ?entity vital-core:vitaltype haley:KGEntity .
            ?entity haley:hasKGEntityType {self.utils.build_uri_reference(entity_type_uri)} .
            """
        else:
            type_filter = """
            ?entity vital-core:vitaltype haley:KGEntity .
            """
        
        # Build search filter
        search_filter = ""
        if search:
            escaped_search = self.utils.escape_sparql_string(search)
            search_filter = f"""
            ?entity vital:hasName ?name .
            FILTER(CONTAINS(LCASE(?name), LCASE("{escaped_search}")))
            """
        
        # Build pagination
        pagination = self.utils.build_pagination_clause(page_size, offset)
        
        return f"""
        {self.utils.build_prefixes()}
        
        SELECT DISTINCT ?entity WHERE {{
            {self.utils.build_graph_clause(graph_id)} {{
                {type_filter}
                {search_filter}
            }}
        }}
        ORDER BY ?entity
        {pagination}
        """


"""
Mock implementation of KGQueriesEndpoint for entity-to-entity connection queries.

This endpoint provides two distinct query types:
1. Relation-based queries: Find entities connected via Edge_hasKGRelation
2. Frame-based queries: Find entities connected via shared KGFrames

Each query type is completely separate and never combined.
Results focus on connection triples: source entity -> connection -> destination entity
"""

from typing import Dict, Any, Optional, List
import logging
from .mock_base_endpoint import MockBaseEndpoint
from vitalgraph.model.kgqueries_model import (
    KGQueryRequest, KGQueryResponse, RelationConnection, FrameConnection,
    KGQueryStatsResponse
)
from vitalgraph.sparql.kg_connection_query_builder import KGConnectionQueryBuilder

logger = logging.getLogger(__name__)


class MockKGQueriesEndpoint(MockBaseEndpoint):
    """Mock implementation of KGQueriesEndpoint with VitalSigns native functionality."""
    
    def __init__(self, client=None, space_manager=None, *, config=None):
        """Initialize with connection query builder."""
        super().__init__(client, space_manager, config=config)
        self.connection_query_builder = KGConnectionQueryBuilder()
        self.haley_prefix = "http://vital.ai/ontology/haley-ai-kg#"
    
    def query_connections(self, space_id: str, graph_id: str, 
                         query_request: KGQueryRequest) -> KGQueryResponse:
        """
        Query entities connected via relations or shared frames based on query_type.
        
        Args:
            space_id: Space identifier
            graph_id: Graph identifier
            query_request: Query criteria and pagination
            
        Returns:
            KGQueryResponse with connections based on query_type
        """
        query_type = query_request.criteria.query_type
        logger.info(f"Executing {query_type} query in space {space_id}, graph {graph_id}")
        
        try:
            # Get space and validate
            space = self.space_manager.get_space(space_id) if self.space_manager else None
            if not space:
                return KGQueryResponse(
                    query_type=query_type,
                    relation_connections=[] if query_type == "relation" else None,
                    frame_connections=[] if query_type == "frame" else None,
                    total_count=0,
                    page_size=query_request.page_size,
                    offset=query_request.offset
                )
            
            if query_type == "relation":
                return self._execute_relation_query(space, graph_id, query_request)
            elif query_type == "frame":
                return self._execute_frame_query(space, graph_id, query_request)
            else:
                raise ValueError(f"Invalid query_type: {query_type}. Must be 'relation' or 'frame'")
                
        except Exception as e:
            logger.error(f"Error executing {query_type} query: {e}")
            return KGQueryResponse(
                query_type=query_type,
                relation_connections=[] if query_type == "relation" else None,
                frame_connections=[] if query_type == "frame" else None,
                total_count=0,
                page_size=query_request.page_size,
                offset=query_request.offset
            )
    
    def _execute_relation_query(self, space, graph_id: str, query_request: KGQueryRequest) -> KGQueryResponse:
        """Execute relation-based query."""
        # Build SPARQL query for relation connections
        sparql_query = self.connection_query_builder.build_relation_query(
            query_request.criteria, graph_id
        )
        
        logger.debug(f"Generated relation SPARQL query: {sparql_query}")
        
        # Execute SPARQL query
        results = space.triple_store.query(sparql_query)
        
        # Convert results to RelationConnection objects
        connections = []
        for result in results:
            connection = RelationConnection(
                source_entity_uri=str(result['source_entity']),
                destination_entity_uri=str(result['destination_entity']),
                relation_edge_uri=str(result['relation_edge']),
                relation_type_uri=str(result['relation_type'])
            )
            connections.append(connection)
        
        # Apply pagination
        total_count = len(connections)
        start_idx = query_request.offset
        end_idx = start_idx + query_request.page_size
        paginated_connections = connections[start_idx:end_idx]
        
        logger.info(f"Found {total_count} relation connections, returning {len(paginated_connections)}")
        
        return KGQueryResponse(
            query_type="relation",
            relation_connections=paginated_connections,
            frame_connections=None,
            total_count=total_count,
            page_size=query_request.page_size,
            offset=query_request.offset
        )
    
    def _execute_frame_query(self, space, graph_id: str, query_request: KGQueryRequest) -> KGQueryResponse:
        """Execute frame-based query."""
        # Build SPARQL query for frame connections
        sparql_query = self.connection_query_builder.build_frame_query(
            query_request.criteria, graph_id
        )
        
        logger.debug(f"Generated frame SPARQL query: {sparql_query}")
        
        # Execute SPARQL query
        results = space.triple_store.query(sparql_query)
        
        # Convert results to FrameConnection objects
        connections = []
        for result in results:
            connection = FrameConnection(
                source_entity_uri=str(result['source_entity']),
                destination_entity_uri=str(result['destination_entity']),
                shared_frame_uri=str(result['shared_frame']),
                frame_type_uri=str(result['frame_type'])
            )
            connections.append(connection)
        
        # Apply pagination
        total_count = len(connections)
        start_idx = query_request.offset
        end_idx = start_idx + query_request.page_size
        paginated_connections = connections[start_idx:end_idx]
        
        logger.info(f"Found {total_count} frame connections, returning {len(paginated_connections)}")
        
        return KGQueryResponse(
            query_type="frame",
            relation_connections=None,
            frame_connections=paginated_connections,
            total_count=total_count,
            page_size=query_request.page_size,
            offset=query_request.offset
        )
    
    def get_query_stats(self, space_id: str, graph_id: str) -> KGQueryStatsResponse:
        """
        Get statistics about entities, relations, and frames in the graph.
        
        Args:
            space_id: Space identifier
            graph_id: Graph identifier
            
        Returns:
            KGQueryStatsResponse with graph statistics
        """
        logger.info(f"Getting query stats for space {space_id}, graph {graph_id}")
        
        try:
            # Get space and validate
            space = self.space_manager.get_space(space_id) if self.space_manager else None
            if not space:
                return KGQueryStatsResponse(
                    total_entities=0,
                    total_relations=0,
                    total_frames=0,
                    relation_connections_count=0,
                    frame_connections_count=0
                )
            
            # Count entities
            entity_count_query = f"""
            PREFIX haley: <{self.haley_prefix}>
            SELECT (COUNT(DISTINCT ?entity) as ?count) WHERE {{
                GRAPH <{graph_id}> {{
                    ?entity a haley:KGEntity .
                }}
            }}
            """
            entity_results = space.triple_store.query(entity_count_query)
            total_entities = int(entity_results[0]['count']) if entity_results else 0
            
            # Count relations
            relation_count_query = f"""
            PREFIX haley: <{self.haley_prefix}>
            SELECT (COUNT(DISTINCT ?relation) as ?count) WHERE {{
                GRAPH <{graph_id}> {{
                    ?relation a haley:Edge_hasKGRelation .
                }}
            }}
            """
            relation_results = space.triple_store.query(relation_count_query)
            total_relations = int(relation_results[0]['count']) if relation_results else 0
            
            # Count frames
            frame_count_query = f"""
            PREFIX haley: <{self.haley_prefix}>
            SELECT (COUNT(DISTINCT ?frame) as ?count) WHERE {{
                GRAPH <{graph_id}> {{
                    ?frame a haley:KGFrame .
                }}
            }}
            """
            frame_results = space.triple_store.query(frame_count_query)
            total_frames = int(frame_results[0]['count']) if frame_results else 0
            
            # Count relation connections (entity pairs connected via relations)
            relation_connections_query = f"""
            PREFIX haley: <{self.haley_prefix}>
            PREFIX vital: <http://vital.ai/ontology/vital-core#>
            SELECT (COUNT(*) as ?count) WHERE {{
                GRAPH <{graph_id}> {{
                    ?relation a haley:Edge_hasKGRelation .
                    ?relation vital:hasEdgeSource ?source .
                    ?relation vital:hasEdgeDestination ?dest .
                }}
            }}
            """
            relation_conn_results = space.triple_store.query(relation_connections_query)
            relation_connections_count = int(relation_conn_results[0]['count']) if relation_conn_results else 0
            
            # Count frame connections (entity pairs connected via shared frames)
            frame_connections_query = f"""
            PREFIX haley: <{self.haley_prefix}>
            PREFIX vital: <http://vital.ai/ontology/vital-core#>
            SELECT (COUNT(*) as ?count) WHERE {{
                GRAPH <{graph_id}> {{
                    ?edge1 a haley:Edge_hasEntityKGFrame .
                    ?edge1 vital:hasEdgeSource ?source .
                    ?edge1 vital:hasEdgeDestination ?frame .
                    
                    ?edge2 a haley:Edge_hasEntityKGFrame .
                    ?edge2 vital:hasEdgeSource ?dest .
                    ?edge2 vital:hasEdgeDestination ?frame .
                    
                    FILTER(?source != ?dest)
                }}
            }}
            """
            frame_conn_results = space.triple_store.query(frame_connections_query)
            frame_connections_count = int(frame_conn_results[0]['count']) if frame_conn_results else 0
            
            logger.info(f"Stats: {total_entities} entities, {total_relations} relations, {total_frames} frames")
            
            return KGQueryStatsResponse(
                total_entities=total_entities,
                total_relations=total_relations,
                total_frames=total_frames,
                relation_connections_count=relation_connections_count,
                frame_connections_count=frame_connections_count
            )
            
        except Exception as e:
            logger.error(f"Error getting query stats: {e}")
            return KGQueryStatsResponse(
                total_entities=0,
                total_relations=0,
                total_frames=0,
                relation_connections_count=0,
                frame_connections_count=0
            )



"""
Mock implementation of KGEntitiesEndpoint for testing with VitalSigns native JSON-LD functionality.

This implementation uses:
- VitalSigns native object creation and conversion
- pyoxigraph in-memory SPARQL quad store for data persistence
- Proper vitaltype handling for KGEntity objects
- Complete CRUD operations following real endpoint patterns
"""

from typing import Dict, Any, Optional, List
from .mock_base_endpoint import MockBaseEndpoint
from vitalgraph.model.kgentities_model import (
    EntitiesResponse, EntityCreateResponse, EntityUpdateResponse, EntityDeleteResponse,
    EntityQueryRequest, EntityQueryResponse, EntityGraphResponse, EntityGraphDeleteResponse,
    EntitiesGraphResponse
)
from vitalgraph.model.kgframes_model import FrameCreateResponse, FrameUpdateResponse, FrameDeleteResponse
from vitalgraph.model.jsonld_model import JsonLdDocument
from vitalgraph.sparql.grouping_uri_queries import GroupingURIQueryBuilder, GroupingURIGraphRetriever
from vitalgraph.sparql.graph_validation import EntityGraphValidator
from ai_haley_kg_domain.model.KGEntity import KGEntity
from ai_haley_kg_domain.model.KGFrame import KGFrame
from ai_haley_kg_domain.model.KGSlot import KGSlot
from vital_ai_vitalsigns.model.VITAL_Edge import VITAL_Edge


class MockKGEntitiesEndpoint(MockBaseEndpoint):
    """Mock implementation of KGEntitiesEndpoint with VitalSigns native functionality."""
    
    def __init__(self, client=None, space_manager=None, *, config=None):
        """Initialize with SPARQL grouping URI functionality and graph validators."""
        super().__init__(client, space_manager, config=config)
        self.grouping_uri_builder = GroupingURIQueryBuilder()
        self.graph_retriever = GroupingURIGraphRetriever(self.grouping_uri_builder)
        self.entity_validator = EntityGraphValidator()
        self.haley_prefix = "http://vital.ai/ontology/haley-ai-kg#"
        
        # Initialize relations sub-endpoint
        from .mock_kgrelations_endpoint import MockKGRelationsEndpoint
        self.relations = MockKGRelationsEndpoint(client, space_manager, config=config)
    
    def list_kgentities(self, space_id: str, graph_id: str, page_size: int = 10, offset: int = 0, search: Optional[str] = None, include_entity_graph: bool = False) -> EntitiesResponse:
        """
        List KGEntities with pagination and optional search using pyoxigraph SPARQL queries.
        
        Args:
            space_id: Space identifier
            graph_id: Graph identifier
            page_size: Number of entities per page
            offset: Offset for pagination
            search: Optional search term
            include_entity_graph: If True, include complete entity graphs with frames and slots
            
        Returns:
            EntitiesResponse with VitalSigns native JSON-LD document
        """
        from vitalgraph.kg.kgentity_list_endpoint_impl import list_kgentities_impl
        return list_kgentities_impl(self, space_id, graph_id, page_size, offset, search, include_entity_graph)
    
    def get_kgentity(self, space_id: str, graph_id: str, uri: str, include_entity_graph: bool = False) -> EntityGraphResponse:
        """
        Get a specific KGEntity by URI with optional complete graph using grouping URIs.
        
        Args:
            space_id: Space identifier
            graph_id: Graph identifier
            uri: Entity URI
            include_entity_graph: If True, include complete entity graph using hasKGGraphURI
            
        Returns:
            EntityGraphResponse with entity and optional complete graph
        """
        from vitalgraph.kg.kgentity_get_endpoint_impl import get_kgentity_impl
        return get_kgentity_impl(self, space_id, graph_id, uri, include_entity_graph)
    
    def create_kgentities(self, space_id: str, graph_id: str, document: JsonLdDocument) -> EntityCreateResponse:
        """Create KGEntities from JSON-LD document with grouping URI enforcement."""
        from vitalgraph.kg.kgentity_create_endpoint_impl import create_kgentities_impl
        return create_kgentities_impl(self, space_id, graph_id, document)
    
    def update_kgentities(self, space_id: str, graph_id: str, document: JsonLdDocument, 
                         operation_mode: str = "update", parent_uri: str = None) -> EntityUpdateResponse:
        """
        Update KGEntities with proper entity lifecycle management.
        
        This method implements the complete entity update requirements:
        - Parent object URI validation (if provided)
        - Complete entity graph structure validation (entity→frame, frame→frame, frame→slot)
        - URI set matching validation for updates
        - Proper structure verification
        - Atomic operations with rollback
        
        Args:
            space_id: Space identifier
            graph_id: Graph identifier  
            document: JsonLdDocument containing complete entity graph structure
            operation_mode: "create", "update", or "upsert"
            parent_uri: Optional parent object URI (entity or parent frame)
            
        Returns:
            EntityUpdateResponse with updated URI and operation details
        """
        from vitalgraph.kg.kgentity_update_endpoint_impl import update_kgentities_impl
        return update_kgentities_impl(self, space_id, graph_id, document, operation_mode, parent_uri)
    
    def delete_kgentity(self, space_id: str, graph_id: str, uri: str, delete_entity_graph: bool = False) -> EntityDeleteResponse:
        """Delete a KGEntity by URI using pyoxigraph SPARQL DELETE."""
        from vitalgraph.kg.kgentity_delete_endpoint_impl import delete_kgentity_impl
        return delete_kgentity_impl(self, space_id, graph_id, uri, delete_entity_graph)
    
    def delete_kgentities_batch(self, space_id: str, graph_id: str, uri_list: str) -> EntityDeleteResponse:
        """
        Delete multiple KGEntities by URI list using pyoxigraph batch operations.
        
        Args:
            space_id: Space identifier
            graph_id: Graph identifier
            uri_list: Comma-separated list of URIs to delete
            
        Returns:
            EntityDeleteResponse with total deletion count
        """
        from vitalgraph.kg.kgentity_delete_endpoint_impl import delete_kgentities_batch_impl
        return delete_kgentities_batch_impl(self, space_id, graph_id, uri_list)
    
    def get_kgentity_frames(self, space_id: str, graph_id: str, entity_uri: Optional[str] = None, 
                           page_size: int = 10, offset: int = 0, search: Optional[str] = None) -> Dict[str, Any]:
        """
        Get frames associated with KGEntities using pyoxigraph SPARQL queries.
        
        Args:
            space_id: Space identifier
            graph_id: Graph identifier
            entity_uri: Optional specific entity URI
            page_size: Number of frames per page
            offset: Offset for pagination
            search: Optional search term
            
        Returns:
            Dictionary with entity frames data
        """
        from vitalgraph.kg.kgentity_get_endpoint_impl import get_kgentity_frames_impl
        return get_kgentity_frames_impl(self, space_id, graph_id, entity_uri, page_size, offset, search)
    
    def query_entities(self, space_id: str, graph_id: str, query_request: EntityQueryRequest) -> EntityQueryResponse:
        """
        Query KGEntities using criteria-based search.
        
        Args:
            space_id: Space identifier
            graph_id: Graph identifier
            query_request: EntityQueryRequest containing search criteria and pagination
            
        Returns:
            EntityQueryResponse containing list of matching entity URIs and pagination info
        """
        from vitalgraph.kg.kgentity_query_endpoint_impl import query_entities_impl
        return query_entities_impl(self, space_id, graph_id, query_request)
    
    def list_kgentities_with_graphs(self, space_id: str, graph_id: str, page_size: int = 10, offset: int = 0,
                                   search: Optional[str] = None, include_entity_graphs: bool = False) -> EntitiesGraphResponse:
        """
        List KGEntities with optional complete graphs.
        
        Args:
            space_id: Space identifier
            graph_id: Graph identifier
            page_size: Number of items per page
            offset: Offset for pagination
            search: Optional search term
            include_entity_graphs: If True, include complete entity graphs for all entities
            
        Returns:
            EntitiesGraphResponse containing entities and optional complete graphs
        """
        from vitalgraph.kg.kgentity_list_endpoint_impl import list_kgentities_with_graphs_impl
        return list_kgentities_with_graphs_impl(self, space_id, graph_id, page_size, offset, search, include_entity_graphs)
    
    # New Entity-Frame Relationship Methods (Phase 1A)
    
    def create_entity_frames(self, space_id: str, graph_id: str, entity_uri: str, document: JsonLdDocument, operation_mode: str = "create") -> FrameCreateResponse:
        """Create frames within entity context using Edge_hasKGFrame relationships."""
        from vitalgraph.kg.kgentity_create_endpoint_impl import create_entity_frames_complex_impl
        return create_entity_frames_complex_impl(self, space_id, graph_id, entity_uri, document, operation_mode)
    
    def update_entity_frames(self, space_id: str, graph_id: str, entity_uri: str, document: JsonLdDocument) -> FrameUpdateResponse:
        """
        Update frames within entity context using Edge_hasKGFrame relationships.
        
        Args:
            space_id: Space identifier
            graph_id: Graph identifier
            entity_uri: Entity URI to update frames for
            document: JSON-LD document containing updated frames
            
        Returns:
            FrameUpdateResponse with update details
        """
        from vitalgraph.kg.kgentity_update_endpoint_impl import update_entity_frames_impl
        return update_entity_frames_impl(self, space_id, graph_id, entity_uri, document)
    
    def delete_entity_frames(self, space_id: str, graph_id: str, entity_uri: str, frame_uris: List[str]) -> FrameDeleteResponse:
        """Delete frames within entity context using Edge_hasKGFrame relationships."""
        from vitalgraph.kg.kgentity_delete_endpoint_impl import delete_entity_frames_complex_impl
        return delete_entity_frames_complex_impl(self, space_id, graph_id, entity_uri, frame_uris)
    
    # Helper methods for entity graph lifecycle management
    
    def _validate_parent_object(self, space, parent_uri: str, graph_id: str) -> bool:
        """Validate that parent object exists (entity or frame)."""
        from vitalgraph.utils.endpoint_validation import validate_parent_object
        return validate_parent_object(space, parent_uri, graph_id, self.logger)
    
    def _validate_entity_graph_structure(self, objects: list) -> dict:
        """Validate that objects form a complete entity graph structure."""
        from vitalgraph.utils.validation_utils import validate_entity_graph_structure
        return validate_entity_graph_structure(objects)
    
    def _handle_entity_create_mode(self, space, graph_id: str, entity_uri: str, incoming_objects: list, 
                                  incoming_uris: set, parent_uri: str = None) -> EntityUpdateResponse:
        """Handle CREATE mode: verify none of the objects already exist."""
        from vitalgraph.kg.kgentity_create_endpoint_impl import handle_entity_create_mode_impl
        return handle_entity_create_mode_impl(self, space, graph_id, entity_uri, incoming_objects, incoming_uris, parent_uri)
    
    def _handle_entity_update_mode(self, space, graph_id: str, entity_uri: str, incoming_objects: list,
                                  incoming_uris: set, parent_uri: str = None) -> EntityUpdateResponse:
        """Handle UPDATE mode: verify entity exists and replace with new content."""
        from vitalgraph.kg.kgentity_update_endpoint_impl import handle_entity_update_mode_impl
        return handle_entity_update_mode_impl(self, space, graph_id, entity_uri, incoming_objects, incoming_uris, parent_uri)
    
    def _handle_entity_upsert_mode(self, space, graph_id: str, entity_uri: str, incoming_objects: list,
                                  incoming_uris: set, parent_uri: str = None) -> EntityUpdateResponse:
        """Handle UPSERT mode: create or update, verify structure and entity URI consistency."""
        from vitalgraph.kg.kgentity_upsert_endpoint_impl import handle_entity_upsert_mode_impl
        return handle_entity_upsert_mode_impl(self, space, graph_id, entity_uri, incoming_objects, incoming_uris, parent_uri)
    
    def _object_exists_in_store(self, space, uri: str, graph_id: str) -> bool:
        """Check if any object with the given URI exists in the store."""
        from vitalgraph.utils.sparql_helpers import check_object_exists_in_graph
        return check_object_exists_in_graph(space, uri, graph_id)
    
    def _entity_exists_in_store(self, space, entity_uri: str, graph_id: str) -> bool:
        """Check if entity exists in the store."""
        try:
            if graph_id:
                query = f"""
                PREFIX haley: <http://vital.ai/ontology/haley-ai-kg#>
                ASK {{
                    GRAPH <{graph_id}> {{
                        <{entity_uri}> a haley:KGEntity .
                    }}
                }}
                """
            else:
                query = f"""
                PREFIX haley: <http://vital.ai/ontology/haley-ai-kg#>
                ASK {{
                    <{entity_uri}> a haley:KGEntity .
                }}
                """
            return space.store.query(query)
        except Exception:
            return False
    
    def _get_current_entity_objects(self, space, entity_uri: str, graph_id: str) -> list:
        """Get all current objects belonging to an entity via grouping URIs."""
        from vitalgraph.kg.kgentity_get_endpoint_impl import get_current_entity_objects_impl
        return get_current_entity_objects_impl(self, space, entity_uri, graph_id)
    
    def _validate_parent_connection(self, space, parent_uri: str, entity_uri: str, graph_id: str, incoming_objects: list) -> bool:
        """Validate that there's a proper connection between parent and entity in the incoming objects."""
        from vitalgraph.utils.endpoint_validation import validate_parent_connection
        return validate_parent_connection(space, parent_uri, entity_uri, graph_id, incoming_objects, 
                                        "Edge_hasEntityKGFrame", self.logger)
    
    def _backup_entity_graph(self, space, entity_uri: str, graph_id: str) -> dict:
        """Backup entity graph for rollback capability."""
        try:
            # This is a simplified implementation
            # In practice, would need to backup all triples for the entity graph
            return {"entity_uri": entity_uri, "backup_created": True}
        except Exception as e:
            self.logger.error(f"Error backing up entity graph: {e}")
            return {}
    
    def _restore_entity_graph_from_backup(self, space, entity_uri: str, graph_id: str, backup_data: dict) -> bool:
        """Restore entity graph from backup."""
        try:
            # This is a simplified implementation
            # In practice, would restore all backed up triples
            self.logger.info(f"Entity graph backup restore for {entity_uri} - simplified implementation")
            return True
        except Exception as e:
            self.logger.error(f"Error restoring entity graph from backup: {e}")
            return False

    # Helper methods for SPARQL-based graph retrieval
    
    def _object_to_triples(self, obj, graph_id: str) -> List[tuple]:
        """Convert a single VitalSigns object to RDF triples."""
        try:
            # Use base class method to convert objects to triples
            triples = self._convert_objects_to_triples([obj])
            # Add graph context to each triple
            quads = []
            for triple in triples:
                if len(triple) == 3:
                    # Add graph_id as fourth element to make it a quad
                    quads.append((triple[0], triple[1], triple[2], graph_id))
                else:
                    quads.append(triple)
            return quads
        except Exception as e:
            self.logger.error(f"Error converting object to triples: {e}")
            return []
    
    def _store_triples(self, space, triples: List[tuple]) -> bool:
        """Store RDF triples/quads in the pyoxigraph store."""
        try:
            # Use base class method to insert quads
            return self._insert_quads_to_store(space, triples)
        except Exception as e:
            self.logger.error(f"Error storing triples: {e}")
            return False
    
    def _generate_uuid(self) -> str:
        """Generate a UUID for new objects."""
        from vitalgraph.utils.vitalsigns_helpers import generate_uuid
        return generate_uuid()
    
    def _execute_sparql_update(self, space, update_query: str) -> bool:
        """
        Execute SPARQL update query using pyoxigraph capabilities.
        
        Args:
            space: Mock space instance with pyoxigraph
            update_query: SPARQL update query string
            
        Returns:
            True if successful, False otherwise
        """
        try:
            space.update_sparql(update_query)
            return True
            
        except Exception as e:
            self.logger.error(f"Error executing SPARQL update: {e}")
            return False
    
    def _get_single_entity(self, space, graph_id: str, entity_uri: str) -> EntityGraphResponse:
        """Get just the entity itself (standard retrieval)."""
        from vitalgraph.kg.kgentity_get_endpoint_impl import get_single_entity_impl
        return get_single_entity_impl(self, space, graph_id, entity_uri)
    
    def _get_entity_with_complete_graph(self, space, graph_id: str, entity_uri: str) -> EntityGraphResponse:
        """Get entity with complete graph using hasKGGraphURI."""
        from vitalgraph.kg.kgentity_get_endpoint_impl import get_entity_with_complete_graph_impl
        return get_entity_with_complete_graph_impl(self, space, graph_id, entity_uri)
    
    def _convert_triples_to_jsonld(self, triples: List[Dict[str, str]]) -> Dict[str, Any]:
        """Convert list of triples to JSON-LD document."""
        from vitalgraph.utils.vitalsigns_helpers import convert_triples_to_jsonld
        return convert_triples_to_jsonld(triples, self.logger)
    
    def _convert_triples_to_vitalsigns_objects(self, triples: List[Dict[str, str]]) -> List[Any]:
        """Convert triples to VitalSigns objects using list-specific RDF functions."""
        from vitalgraph.utils.vitalsigns_helpers import convert_triples_to_vitalsigns_objects
        return convert_triples_to_vitalsigns_objects(triples, self.logger)
    
    def _simple_triples_to_jsonld(self, triples: List[Dict[str, str]]) -> Dict[str, Any]:
        """Fallback simple conversion of triples to JSON-LD."""
        from vitalgraph.utils.vitalsigns_helpers import simple_triples_to_jsonld
        return simple_triples_to_jsonld(triples)
    
    def _strip_grouping_uris(self, document: JsonLdDocument) -> JsonLdDocument:
        """Strip any existing hasKGGraphURI values from client document."""
        from vitalgraph.utils.vitalsigns_helpers import strip_grouping_uris_from_document
        return strip_grouping_uris_from_document(document)
    
    def _set_entity_grouping_uris(self, objects: List[Any], entity_uri: str) -> None:
        """Set hasKGGraphURI to entity_uri for all entity graph components."""
        from vitalgraph.utils.graph_operations import set_entity_grouping_uris
        set_entity_grouping_uris(objects, entity_uri, self.logger)
    
    def _set_dual_grouping_uris(self, objects: List[Any], entity_uri: str) -> None:
        """Set both entity-level and frame-level grouping URIs for proper graph retrieval."""
        from vitalgraph.utils.graph_operations import set_dual_grouping_uris
        set_dual_grouping_uris(objects, entity_uri, self.logger)
    
    def _process_complete_entity_document(self, document: JsonLdDocument, entity_uri: str) -> List[Any]:
        """Process complete entity document to extract and validate all KG objects using VitalSigns."""
        from vitalgraph.kg.kgentity_create_endpoint_impl import process_complete_entity_document_impl
        return process_complete_entity_document_impl(self, document, entity_uri)
    
    def _create_vitalsigns_objects_from_jsonld(self, jsonld_document: Dict[str, Any]) -> List[Any]:
        """Create VitalSigns objects from JSON-LD document using VitalSigns native methods."""
        from vitalgraph.utils.vitalsigns_helpers import create_vitalsigns_objects_from_jsonld
        return create_vitalsigns_objects_from_jsonld(jsonld_document, self.logger)
    
    def _jsonld_to_triples(self, jsonld_document: Dict[str, Any]) -> List[Dict[str, str]]:
        """
        Convert JSON-LD document to triples using pyoxigraph.
        
        Args:
            jsonld_document: JSON-LD document dict
            
        Returns:
            List of triple dicts with 'subject', 'predicate', 'object' keys
        """
        from vitalgraph.utils.vitalsigns_conversion_utils import jsonld_to_triples_impl
        return jsonld_to_triples_impl(self, jsonld_document)
    
    def _delete_entity_graph_from_store(self, space, entity_uri: str, graph_id: str) -> bool:
        """
        Delete complete entity graph using hasKGGraphURI grouping property.
        
        Args:
            space: Space object
            entity_uri: URI of the entity whose graph should be deleted
            graph_id: Graph ID
            
        Returns:
            True if deletion was successful
        """
        try:
            # Use grouping URI query to find all subjects with hasKGGraphURI = entity_uri
            query = self.grouping_uri_builder.build_entity_graph_subjects_query(entity_uri, graph_id)
            results = self._execute_sparql_query(space, query)
            
            # Delete all subjects that belong to this entity graph
            deleted_count = 0
            if results.get("bindings"):
                for binding in results["bindings"]:
                    subject_uri = binding.get("subject", {}).get("value", "")
                    if subject_uri:
                        if self._delete_quads_from_store(space, subject_uri, graph_id):
                            deleted_count += 1
            
            self.logger.debug(f"Deleted {deleted_count} subjects from entity graph {entity_uri}")
            return deleted_count > 0
            
        except Exception as e:
            self.logger.error(f"Failed to delete entity graph {entity_uri}: {e}")
            return False
    
    def _build_entity_query_from_criteria(self, criteria, graph_id: str, page_size: int, offset: int) -> str:
        """Build SPARQL query from EntityQueryCriteria."""
        
        # Start with basic entity query
        query_parts = [
            f"SELECT DISTINCT ?entity WHERE {{",
            f"  GRAPH <{graph_id}> {{",
            f"    ?entity a <{self.haley_prefix}KGEntity> ."
        ]
        
        # Add search string filter if provided
        if criteria.search_string:
            query_parts.extend([
                f"    ?entity ?searchProp ?searchValue .",
                f"    FILTER(CONTAINS(LCASE(STR(?searchValue)), LCASE(\"{criteria.search_string}\")))"
            ])
        
        # Add entity type filter if provided
        if criteria.entity_type:
            query_parts.append(f"    ?entity a <{criteria.entity_type}> .")
        
        # Add frame type filter if provided
        if criteria.frame_type:
            query_parts.extend([
                f"    ?edge_frame a <{self.haley_prefix}Edge_hasKGFrame> .",
                f"    ?edge_frame <http://vital.ai/ontology/vital-core#hasEdgeSource> ?entity .",
                f"    ?edge_frame <http://vital.ai/ontology/vital-core#hasEdgeDestination> ?frame .",
                f"    ?frame a <{criteria.frame_type}> ."
            ])
        
        # Add slot criteria filters if provided
        if criteria.slot_criteria:
            for i, slot_criterion in enumerate(criteria.slot_criteria):
                slot_var = f"slot{i}"
                query_parts.extend([
                    f"    ?edge_frame{i} a <{self.haley_prefix}Edge_hasKGFrame> .",
                    f"    ?edge_frame{i} <http://vital.ai/ontology/vital-core#hasEdgeSource> ?entity .",
                    f"    ?edge_frame{i} <http://vital.ai/ontology/vital-core#hasEdgeDestination> ?frame{i} .",
                    f"    ?edge_slot{i} a <{self.haley_prefix}Edge_hasKGSlot> .",
                    f"    ?edge_slot{i} <http://vital.ai/ontology/vital-core#hasEdgeSource> ?frame{i} .",
                    f"    ?edge_slot{i} <http://vital.ai/ontology/vital-core#hasEdgeDestination> ?{slot_var} .",
                    f"    ?{slot_var} a <{slot_criterion.slot_type}> ."
                ])
                
                # Add value comparison if provided
                if slot_criterion.value and slot_criterion.comparator != "exists":
                    if slot_criterion.comparator == "contains":
                        query_parts.extend([
                            f"    ?{slot_var} ?slotProp{i} ?slotValue{i} .",
                            f"    FILTER(CONTAINS(LCASE(STR(?slotValue{i})), LCASE(\"{slot_criterion.value}\")))"
                        ])
                    else:
                        # For other comparators, we'd need more sophisticated handling
                        query_parts.extend([
                            f"    ?{slot_var} ?slotProp{i} \"{slot_criterion.value}\" ."
                        ])
        
        # Close the query
        query_parts.extend([
            "  }",
            "}",
            f"LIMIT {page_size}",
            f"OFFSET {offset}"
        ])
        
        return "\n".join(query_parts)
    
    # ========================================
    # Sub-Endpoint Operations: /kgentities/kgframes
    # ========================================
    
    
    def get_entity_frames(self, space_id: str, graph_id: str, entity_uri: str) -> JsonLdDocument:
        """Get frames for an entity using Edge_hasKGFrame relationships."""
        from vitalgraph.kg.kgentity_get_endpoint_impl import get_entity_frames_complex_impl
        return get_entity_frames_complex_impl(self, space_id, graph_id, entity_uri)
    
    
    # ========================================
    # Helper Methods for Sub-Endpoint Operations
    # ========================================
    
    def _validate_entity_exists(self, space, graph_id: str, entity_uri: str) -> bool:
        """Check if entity exists in the graph."""
        query = f"""
        ASK {{
            GRAPH <{graph_id}> {{
                <{entity_uri}> a <http://vital.ai/ontology/haley-ai-kg#KGEntity> .
            }}
        }}
        """
        return bool(space.store.query(query))
    
    def _frame_exists(self, space, graph_id: str, frame_uri: str) -> bool:
        """Check if frame exists in the graph."""
        query = f"""
        ASK {{
            GRAPH <{graph_id}> {{
                <{frame_uri}> a <http://vital.ai/ontology/haley-ai-kg#KGFrame> .
            }}
        }}
        """
        return bool(space.store.query(query))
    
    def _validate_entity_frame_connection(self, space, graph_id: str, entity_uri: str, frame_uri: str) -> bool:
        """Check if frame is connected to entity via Edge_hasEntityKGFrame."""
        query = f"""
        ASK {{
            GRAPH <{graph_id}> {{
                ?edge a <http://vital.ai/ontology/haley-ai-kg#Edge_hasEntityKGFrame> ;
                      <http://vital.ai/ontology/vital-core#hasEdgeSource> <{entity_uri}> ;
                      <http://vital.ai/ontology/vital-core#hasEdgeDestination> <{frame_uri}> .
            }}
        }}
        """
        return bool(space.store.query(query))
    
    def _get_entity_frame_edge(self, space, graph_id: str, entity_uri: str, frame_uri: str) -> Any:
        """Get the specific Edge_hasEntityKGFrame object between entity and frame."""
        try:
            # Query for the specific edge
            edge_query = f"""
            PREFIX haley: <http://vital.ai/ontology/haley-ai-kg#>
            PREFIX vital: <http://vital.ai/ontology/vital-core#>
            
            SELECT ?edge ?prop ?value WHERE {{
                GRAPH <{graph_id}> {{
                    ?edge a haley:Edge_hasEntityKGFrame ;
                          vital:hasEdgeSource <{entity_uri}> ;
                          vital:hasEdgeDestination <{frame_uri}> ;
                          ?prop ?value .
                }}
            }}
            """
            
            edge_results = list(space.store.query(edge_query))
            if not edge_results:
                return None
            
            # Convert to triples and then to VitalSigns object
            edge_triples = []
            for result in edge_results:
                edge_uri = str(result['edge']).strip('<>')
                prop_uri = str(result['prop'])
                
                value_obj = result['value']
                if hasattr(value_obj, 'value'):
                    value = str(value_obj.value)
                else:
                    value = str(value_obj)
                
                from vital_ai_vitalsigns.utils.uri_utils import validate_rfc3986
                edge_triples.append({
                    'subject': f'<{edge_uri}>',
                    'predicate': f'<{prop_uri}>',
                    'object': f'<{value}>' if validate_rfc3986(str(value), rule='URI') else f'"{value}"'
                })
            
            # Convert to VitalSigns object
            edge_objects = self._convert_triples_to_vitalsigns_objects(edge_triples)
            return edge_objects[0] if edge_objects else None
            
        except Exception as e:
            self.logger.error(f"Error getting entity-frame edge: {e}")
            return None
    
    def _create_entity_frame_edges(self, objects: List[Any], entity_uri: str, operation_mode: str) -> None:
        """Create Edge_hasEntityKGFrame relationships between entity and frames."""
        from ai_haley_kg_domain.model.KGFrame import KGFrame
        from ai_haley_kg_domain.model.Edge_hasEntityKGFrame import Edge_hasEntityKGFrame
        
        # Find all frames in the objects
        frames = [obj for obj in objects if isinstance(obj, KGFrame)]
        self.logger.info(f"Creating entity-frame edges for {len(frames)} frames")
        
        for frame in frames:
            # Check if edge already exists (for upsert mode)
            edge_exists = any(
                isinstance(obj, Edge_hasEntityKGFrame) and 
                str(obj.edgeSource) == entity_uri and 
                str(obj.edgeDestination) == str(frame.URI)
                for obj in objects
            )
            
            if not edge_exists:
                # Create new edge
                edge = Edge_hasEntityKGFrame()
                edge.URI = f"http://vital.ai/haley.ai/app/Edge_hasEntityKGFrame/{entity_uri.split('/')[-1]}_{frame.URI.split('/')[-1]}_edge"
                edge.edgeSource = entity_uri
                edge.edgeDestination = str(frame.URI)
                
                # Set grouping URIs
                edge.kGGraphURI = entity_uri
                # Note: Edge_hasEntityKGFrame doesn't have frameGraphURI attribute
                
                self.logger.info(f"Created entity-frame edge: {edge.URI} ({entity_uri} -> {frame.URI})")
                objects.append(edge)
            else:
                self.logger.info(f"Entity-frame edge already exists for {entity_uri} -> {frame.URI}")
    
    def _delete_frame_with_components(self, space, graph_id: str, frame_uri: str) -> bool:
        """Delete frame and all its components (slots, edges)."""
        try:
            # Delete all triples where frame is subject
            delete_query = f"""
            DELETE WHERE {{
                GRAPH <{graph_id}> {{
                    <{frame_uri}> ?p ?o .
                }}
            }}
            """
            space.store.update(delete_query)
            
            # Delete all edges pointing to this frame
            delete_edges_query = f"""
            DELETE WHERE {{
                GRAPH <{graph_id}> {{
                    ?edge ?p <{frame_uri}> .
                }}
            }}
            """
            space.store.update(delete_edges_query)
            
            # Delete slots connected to this frame
            delete_slots_query = f"""
            DELETE WHERE {{
                GRAPH <{graph_id}> {{
                    ?edge a <http://vital.ai/ontology/haley-ai-kg#Edge_hasKGSlot> ;
                          <http://vital.ai/ontology/vital-core#hasEdgeSource> <{frame_uri}> ;
                          <http://vital.ai/ontology/vital-core#hasEdgeDestination> ?slot .
                    ?slot ?p ?o .
                    ?edge ?ep ?eo .
                }}
            }}
            """
            space.store.update(delete_slots_query)
            
            return True
            
        except Exception as e:
            self.logger.error(f"Error deleting frame components: {e}")
            return False
    
    def _delete_entity_frame_edge(self, space, graph_id: str, entity_uri: str, frame_uri: str) -> None:
        """Delete Edge_hasEntityKGFrame between entity and frame."""
        try:
            delete_query = f"""
            DELETE WHERE {{
                GRAPH <{graph_id}> {{
                    ?edge a <http://vital.ai/ontology/haley-ai-kg#Edge_hasEntityKGFrame> ;
                          <http://vital.ai/ontology/vital-core#hasEdgeSource> <{entity_uri}> ;
                          <http://vital.ai/ontology/vital-core#hasEdgeDestination> <{frame_uri}> .
                    ?edge ?p ?o .
                }}
            }}
            """
            space.store.update(delete_query)
            
        except Exception as e:
            self.logger.error(f"Error deleting entity-frame edge: {e}")
    
    def _validate_frame_structure(self, objects: List[Any]) -> Dict[str, Any]:
        """Validate frame structure and extract frame URIs."""
        try:
            from ai_haley_kg_domain.model.KGFrame import KGFrame
            
            frame_uris = []
            
            for obj in objects:
                if isinstance(obj, KGFrame):
                    if hasattr(obj, 'URI') and obj.URI:
                        frame_uris.append(str(obj.URI))
            
            if not frame_uris:
                return {'valid': False, 'error': 'No valid frames found', 'frame_uris': []}
            
            return {'valid': True, 'error': None, 'frame_uris': frame_uris}
            
        except Exception as e:
            return {'valid': False, 'error': str(e), 'frame_uris': []}
    
    
    def _delete_frame_triples(self, space, graph_id: str, frame_uri: str) -> bool:
        """Delete all triples for a frame."""
        try:
            delete_query = f"""
            DELETE WHERE {{
                GRAPH <{graph_id}> {{
                    <{frame_uri}> ?p ?o .
                }}
            }}
            """
            space.store.update(delete_query)
            return True
            
        except Exception as e:
            self.logger.error(f"Error deleting frame triples: {e}")
            return False



"""
Mock implementation of GraphsEndpoint for testing with VitalSigns native functionality.

This implementation uses:
- Real pyoxigraph graph operations for CREATE, DROP, CLEAR
- SPARQL queries to get actual graph information and triple counts
- VitalSigns native functionality for data conversions
- No mock data generation - all operations use real pyoxigraph storage
"""

import time
from typing import Dict, Any, Optional, List
from .mock_base_endpoint import MockBaseEndpoint
from vitalgraph.model.sparql_model import (
    GraphInfo, SPARQLGraphRequest, SPARQLGraphResponse
)


class MockGraphsEndpoint(MockBaseEndpoint):
    """Mock implementation of GraphsEndpoint."""
    
    def _parse_sparql_integer(self, value_str: str) -> int:
        """Parse SPARQL integer result that may include XSD datatype."""
        try:
            # Handle XSD datatype format like "0"^^<http://www.w3.org/2001/XMLSchema#integer>
            if "^^" in value_str:
                value_str = value_str.split("^^")[0].strip('"')
            return int(value_str)
        except (ValueError, AttributeError):
            return 0
    
    def list_graphs(self, space_id: str) -> List[GraphInfo]:
        """
        List graphs in a space by examining the MockGraph objects managed by the space.
        
        Args:
            space_id: Space identifier
            
        Returns:
            List of GraphInfo objects with graph metadata and triple counts
        """
        self._log_method_call("list_graphs", space_id=space_id)
        
        try:
            # Get space from space manager
            space = self.space_manager.get_space(space_id) if self.space_manager else None
            if not space:
                return []
            
            # List all MockGraph objects managed by the space
            graphs = []
            for mock_graph in space.graphs.values():
                # Get triple count for this graph's URI from the SPARQL store
                query = f"""
                SELECT (COUNT(*) as ?triple_count) WHERE {{
                    GRAPH <{mock_graph.graph_uri}> {{
                        ?s ?p ?o .
                    }}
                }}
                """
                
                results = self._execute_sparql_query(space, query)
                triple_count = 0
                if results.get("bindings"):
                    triple_count_str = results["bindings"][0].get("triple_count", {}).get("value", "0")
                    triple_count = self._parse_sparql_integer(triple_count_str)
                
                # Use current time as placeholder for created/updated times
                current_time = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
                
                graph_info = GraphInfo(
                    graph_uri=mock_graph.graph_uri,
                    triple_count=triple_count,
                    created_time=current_time,
                    updated_time=current_time
                )
                graphs.append(graph_info)
            
            return graphs
            
        except Exception as e:
            self.logger.error(f"Error listing graphs: {e}")
            return []
    
    def get_graph_info(self, space_id: str, graph_uri: str) -> GraphInfo:
        """
        Get information about a specific graph using real pyoxigraph SPARQL queries.
        
        Args:
            space_id: Space identifier
            graph_uri: Graph URI to get information about
            
        Returns:
            GraphInfo object with real graph data
        """
        self._log_method_call("get_graph_info", space_id=space_id, graph_uri=graph_uri)
        
        try:
            # Get space from space manager
            space = self.space_manager.get_space(space_id) if self.space_manager else None
            if not space:
                # Return empty graph info for non-existent space
                current_time = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
                return GraphInfo(
                    graph_uri=graph_uri,
                    triple_count=0,
                    created_time=current_time,
                    updated_time=current_time
                )
            
            # Query for triple count in specific graph
            query = f"""
            SELECT (COUNT(*) as ?triple_count) WHERE {{
                GRAPH <{graph_uri}> {{
                    ?s ?p ?o .
                }}
            }}
            """
            
            results = self._execute_sparql_query(space, query)
            
            triple_count = 0
            if results.get("bindings"):
                triple_count_str = results["bindings"][0].get("triple_count", {}).get("value", "0")
                triple_count = self._parse_sparql_integer(triple_count_str)
            
            # Use current time as placeholder for created/updated times
            current_time = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
            
            return GraphInfo(
                graph_uri=graph_uri,
                triple_count=triple_count,
                created_time=current_time,
                updated_time=current_time
            )
            
        except Exception as e:
            self.logger.error(f"Error getting graph info for {graph_uri}: {e}")
            current_time = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
            return GraphInfo(
                graph_uri=graph_uri,
                triple_count=0,
                updated_time=current_time
            )
    
    def create_graph(self, space_id: str, graph_uri: str) -> SPARQLGraphResponse:
        """
        Create a new MockGraph object in the space.
        
        Args:
            space_id: Space identifier
            graph_uri: Graph URI to create
            
        Returns:
            SPARQLGraphResponse with operation results
        """
        self._log_method_call("create_graph", space_id=space_id, graph_uri=graph_uri)
        
        try:
            # Get space from space manager
            space = self.space_manager.get_space(space_id) if self.space_manager else None
            if not space:
                return SPARQLGraphResponse(
                    success=False,
                    message=f"Space {space_id} not found",
                    operation_time=0.001
                )
            
            start_time = time.time()
            
            # Check if graph already exists
            existing_graph = space.get_graph_by_uri(graph_uri)
            if existing_graph:
                return SPARQLGraphResponse(
                    success=False,
                    message=f"Graph {graph_uri} already exists",
                    operation_time=time.time() - start_time
                )
            
            # Create MockGraph object in the space
            mock_graph = space.add_graph(graph_uri, name=f"Graph {graph_uri}")
            
            operation_time = time.time() - start_time
            
            return SPARQLGraphResponse(
                success=True,
                message="Graph created successfully",
                operation_time=operation_time
            )
                
        except Exception as e:
            self.logger.error(f"Error creating graph {graph_uri}: {e}")
            return SPARQLGraphResponse(
                success=False,
                operation_time=0.001
            )
    
    def drop_graph(self, space_id: str, graph_uri: str, silent: bool = False) -> SPARQLGraphResponse:
        """
        Drop (delete) a MockGraph object from the space and clear its data.
        
        Args:
            space_id: Space identifier
            graph_uri: Graph URI to drop
            silent: Whether to suppress errors if graph doesn't exist
            
        Returns:
            SPARQLGraphResponse with operation results
        """
        self._log_method_call("drop_graph", space_id=space_id, graph_uri=graph_uri, silent=silent)
        
        try:
            # Get space from space manager
            space = self.space_manager.get_space(space_id) if self.space_manager else None
            if not space:
                return SPARQLGraphResponse(
                    success=False,
                    message=f"Space {space_id} not found",
                    operation_time=0.001
                )
            
            start_time = time.time()
            
            # Find the MockGraph by URI
            mock_graph = space.get_graph_by_uri(graph_uri)
            if not mock_graph:
                if silent:
                    return SPARQLGraphResponse(
                        success=True,
                        message="Graph dropped successfully",
                        operation_time=time.time() - start_time
                    )
                else:
                    return SPARQLGraphResponse(
                        success=False,
                        message=f"The graph <{graph_uri}> does not exist",
                        operation_time=time.time() - start_time
                    )
            
            # Remove the MockGraph object from the space (this also clears its data)
            removed = space.remove_graph(mock_graph.graph_id)
            
            operation_time = time.time() - start_time
            
            if removed:
                return SPARQLGraphResponse(
                    success=True,
                    message="Graph dropped successfully",
                    operation_time=operation_time
                )
            else:
                return SPARQLGraphResponse(
                    success=False,
                    message="Graph drop failed",
                    operation_time=operation_time
                )
                
        except Exception as e:
            self.logger.error(f"Error dropping graph {graph_uri}: {e}")
            return SPARQLGraphResponse(
                success=False,
                message=str(e),
                operation_time=0.001
            )
    
    def clear_graph(self, space_id: str, graph_uri: str) -> SPARQLGraphResponse:
        """
        Clear a graph (remove all triples but keep the MockGraph object).
        
        Args:
            space_id: Space identifier
            graph_uri: Graph URI to clear
            
        Returns:
            SPARQLGraphResponse with operation results
        """
        self._log_method_call("clear_graph", space_id=space_id, graph_uri=graph_uri)
        
        try:
            # Get space from space manager
            space = self.space_manager.get_space(space_id) if self.space_manager else None
            if not space:
                return SPARQLGraphResponse(
                    success=False,
                    message=f"Space {space_id} not found",
                    operation_time=0.001
                )
            
            start_time = time.time()
            
            # Find the MockGraph by URI
            mock_graph = space.get_graph_by_uri(graph_uri)
            if not mock_graph:
                return SPARQLGraphResponse(
                    success=False,
                    message=f"The graph <{graph_uri}> does not exist",
                    operation_time=time.time() - start_time
                )
            
            # Clear all triples from this graph using SPARQL DELETE
            delete_query = f"""
            DELETE {{
                GRAPH <{graph_uri}> {{
                    ?s ?p ?o .
                }}
            }}
            WHERE {{
                GRAPH <{graph_uri}> {{
                    ?s ?p ?o .
                }}
            }}
            """
            
            result = space.update_sparql(delete_query)
            operation_time = time.time() - start_time
            
            # SPARQL DELETE doesn't fail for empty graphs, so we consider it successful
            return SPARQLGraphResponse(
                success=True,
                message="Graph cleared successfully",
                operation_time=operation_time
            )
                
        except Exception as e:
            self.logger.error(f"Error clearing graph {graph_uri}: {e}")
            return SPARQLGraphResponse(
                success=False,
                message=str(e),
                operation_time=0.001
            )
    



"""
KG Relations REST API endpoint for VitalGraph.

This module provides REST API endpoints for managing KG relations using JSON-LD 1.1 format.
KG relations represent direct entity-to-entity relationships in the knowledge graph.
"""

import logging
from typing import Dict, List, Optional, Union, Any
from fastapi import APIRouter, Query, Depends, HTTPException, Body
from pydantic import BaseModel, Field
from enum import Enum

from ..model.jsonld_model import JsonLdDocument, JsonLdObject, JsonLdRequest
from ..model.kgrelations_model import (
    RelationsResponse,
    RelationResponse,
    RelationCreateResponse,
    RelationUpdateResponse,
    RelationUpsertResponse,
    RelationDeleteRequest,
    RelationDeleteResponse,
    RelationQueryRequest,
    RelationQueryResponse
)

# Import implementation modules
from ..kg_impl.kgrelations_create_impl import KGRelationsCreateProcessor, OperationMode as CreateOperationMode
from ..kg_impl.kgrelations_read_impl import KGRelationsReadProcessor
from ..kg_impl.kgrelations_delete_impl import KGRelationsDeleteProcessor
from ..kg_impl.kg_backend_utils import FusekiPostgreSQLBackendAdapter


class OperationMode(str, Enum):
    CREATE = "create"
    UPDATE = "update"
    UPSERT = "upsert"


class KGRelationsEndpoint:
    """REST API endpoint for KG Relations operations."""
    
    def __init__(self, space_manager, auth_dependency):
        """Initialize KG Relations endpoint with space manager and authentication."""
        self.space_manager = space_manager
        self.auth_dependency = auth_dependency
        self.logger = logging.getLogger(__name__)
        self.router = APIRouter()
        self.haley_prefix = "http://vital.ai/ontology/haley-ai-kg#"
        self.vital_prefix = "http://vital.ai/ontology/vital-core#"
        
        # Set up routes
        self._setup_routes()
    
    def _setup_routes(self):
        """Set up FastAPI routes for KG Relations operations."""
        
        @self.router.get("/kgrelations", response_model=Union[RelationsResponse, RelationResponse], tags=["KG Relations"])
        async def list_or_get_relations(
            space_id: str = Query(..., description="Space ID"),
            graph_id: str = Query(..., description="Graph ID"),
            relation_uri: Optional[str] = Query(None, description="Single relation URI to retrieve"),
            entity_source_uri: Optional[str] = Query(None, description="Filter by source entity URI"),
            entity_destination_uri: Optional[str] = Query(None, description="Filter by destination entity URI"),
            relation_type_uri: Optional[str] = Query(None, description="Filter by relation type URI"),
            direction: str = Query("all", description="Direction filter: all, incoming, outgoing"),
            page_size: int = Query(10, ge=1, le=1000, description="Number of relations per page"),
            offset: int = Query(0, ge=0, description="Offset for pagination"),
            current_user: Dict = Depends(self.auth_dependency)
        ):
            """
            List KG Relations with filtering and pagination, or get specific relation by URI.
            
            - If relation_uri is provided: returns single relation
            - Otherwise: returns paginated list of relations with optional filtering
            
            Args:
                space_id: Space identifier
                graph_id: Graph identifier
                relation_uri: Optional single relation URI to retrieve
                entity_source_uri: Filter by source entity URI
                entity_destination_uri: Filter by destination entity URI
                relation_type_uri: Filter by relation type URI
                direction: Direction filter (all, incoming, outgoing)
                page_size: Number of relations per page
                offset: Offset for pagination
                
            Returns:
                RelationResponse (if relation_uri provided) or RelationsResponse (for list)
            """
            if relation_uri:
                return await self._get_relation(space_id, graph_id, relation_uri, current_user)
            else:
                return await self._list_relations(
                    space_id, graph_id, entity_source_uri, entity_destination_uri,
                    relation_type_uri, direction, page_size, offset, current_user
                )
        
        @self.router.post("/kgrelations", response_model=Union[RelationCreateResponse, RelationUpdateResponse, RelationUpsertResponse], tags=["KG Relations"])
        async def create_or_update_relations(
            request: JsonLdRequest,
            space_id: str = Query(..., description="Space ID"),
            graph_id: str = Query(..., description="Graph ID"),
            operation_mode: OperationMode = Query(OperationMode.CREATE, description="Operation mode: create, update, or upsert"),
            current_user: Dict = Depends(self.auth_dependency)
        ):
            """
            Create, update, or upsert KG Relations from JSON-LD document or object.
            Uses discriminated union to automatically handle single objects (JsonLdObject) or multiple objects (JsonLdDocument).
            
            - CREATE: Create new relations (fails if relation URIs already exist)
            - UPDATE: Update existing relations (fails if relation URIs don't exist)
            - UPSERT: Create or update relations (always succeeds)
            """
            return await self._create_or_update_relations(space_id, graph_id, request, operation_mode, current_user)
        
        @self.router.delete("/kgrelations", response_model=RelationDeleteResponse, tags=["KG Relations"])
        async def delete_relations(
            space_id: str = Query(..., description="Space ID"),
            graph_id: str = Query(..., description="Graph ID"),
            request: RelationDeleteRequest = Body(...),
            current_user: Dict = Depends(self.auth_dependency)
        ):
            """
            Delete KG Relations by URIs.
            
            Args:
                request: RelationDeleteRequest containing relation URIs to delete
                space_id: Space identifier
                graph_id: Graph identifier
                
            Returns:
                RelationDeleteResponse with deletion results
            """
            return await self._delete_relations(space_id, graph_id, request, current_user)
        
        @self.router.post("/kgrelations/query", response_model=RelationQueryResponse, tags=["KG Relations"])
        async def query_relations(
            query_request: RelationQueryRequest,
            space_id: str = Query(..., description="Space ID"),
            graph_id: str = Query(..., description="Graph ID"),
            current_user: Dict = Depends(self.auth_dependency)
        ):
            """
            Query KG Relations using complex criteria.
            
            Args:
                query_request: RelationQueryRequest containing search criteria
                space_id: Space identifier
                graph_id: Graph identifier
                
            Returns:
                RelationQueryResponse containing matching relation URIs
            """
            return await self._query_relations(space_id, graph_id, query_request, current_user)
    
    async def _list_relations(self, space_id: str, graph_id: str, entity_source_uri: Optional[str],
                            entity_destination_uri: Optional[str], relation_type_uri: Optional[str],
                            direction: str, page_size: int, offset: int, current_user: Dict) -> RelationsResponse:
        """List KG Relations with filtering and pagination."""
        try:
            self.logger.info(f"Listing KG Relations in space {space_id}, graph {graph_id}")
            
            # Get backend implementation
            space_record = self.space_manager.get_space(space_id)
            if not space_record:
                raise HTTPException(status_code=500, detail=f"Space {space_id} not available")
            
            space_impl = space_record.space_impl
            backend_impl = space_impl.get_db_space_impl()
            if not backend_impl:
                raise HTTPException(status_code=500, detail="Backend implementation not available")
            
            # Create backend adapter
            backend = FusekiPostgreSQLBackendAdapter(backend_impl)
            
            # Create read processor
            read_processor = KGRelationsReadProcessor(backend)
            
            # List relations using processor
            triples, total_count = await read_processor.list_relations(
                space_id, graph_id, entity_source_uri, entity_destination_uri,
                relation_type_uri, direction, page_size, offset
            )
            
            # Convert triples to JSON-LD document
            if triples:
                from vital_ai_vitalsigns.model.GraphObject import GraphObject
                graph_objects = GraphObject.from_triples_list(triples)
                relations_doc = self._relations_to_jsonld_document(graph_objects)
            else:
                relations_doc = JsonLdDocument(**{
                    "@context": {"@vocab": "http://vital.ai/ontology/"},
                    "@graph": []
                })
            
            return RelationsResponse(
                relations=relations_doc,
                total_count=total_count,
                page_size=page_size,
                offset=offset
            )
            
        except HTTPException:
            raise
        except Exception as e:
            self.logger.error(f"Error listing KG Relations: {e}")
            raise HTTPException(
                status_code=500,
                detail=f"Failed to list KG Relations: {str(e)}"
            )
    
    async def _get_relation(self, space_id: str, graph_id: str, relation_uri: str, current_user: Dict) -> RelationResponse:
        """Get a specific KG Relation by URI."""
        try:
            self.logger.info(f"Getting KG Relation {relation_uri} in space {space_id}, graph {graph_id}")
            
            # Get backend implementation
            space_record = self.space_manager.get_space(space_id)
            if not space_record:
                raise HTTPException(status_code=500, detail=f"Space {space_id} not available")
            
            space_impl = space_record.space_impl
            backend_impl = space_impl.get_db_space_impl()
            if not backend_impl:
                raise HTTPException(status_code=500, detail="Backend implementation not available")
            
            # Create backend adapter
            backend = FusekiPostgreSQLBackendAdapter(backend_impl)
            
            # Create read processor
            read_processor = KGRelationsReadProcessor(backend)
            
            # Get relation using processor
            triples = await read_processor.get_relation_by_uri(space_id, graph_id, relation_uri)
            
            # Convert triples to JSON-LD document
            if triples:
                from vital_ai_vitalsigns.model.GraphObject import GraphObject
                graph_objects = GraphObject.from_triples_list(triples)
                relation_doc = self._relations_to_jsonld_document(graph_objects)
            else:
                # Return empty JSON-LD document for not found
                relation_doc = JsonLdDocument(**{
                    "@context": {"@vocab": "http://vital.ai/ontology/"},
                    "@graph": []
                })
            
            return RelationResponse(relation=relation_doc)
            
        except HTTPException:
            raise
        except Exception as e:
            self.logger.error(f"Error getting KG Relation: {e}")
            raise HTTPException(
                status_code=500,
                detail=f"Failed to get KG Relation: {str(e)}"
            )
    
    async def _create_or_update_relations(self, space_id: str, graph_id: str, request: JsonLdRequest,
                                        operation_mode: OperationMode, current_user: Dict) -> Union[RelationCreateResponse, RelationUpdateResponse, RelationUpsertResponse]:
        """Create, update, or upsert KG Relations."""
        try:
            self.logger.info(f"Processing KG Relations in space {space_id}, graph {graph_id}, mode '{operation_mode.value}'")
            
            # Get backend implementation
            space_record = self.space_manager.get_space(space_id)
            if not space_record:
                raise HTTPException(status_code=500, detail=f"Space {space_id} not available")
            
            space_impl = space_record.space_impl
            backend_impl = space_impl.get_db_space_impl()
            if not backend_impl:
                raise HTTPException(status_code=500, detail="Backend implementation not available")
            
            # Create backend adapter
            backend = FusekiPostgreSQLBackendAdapter(backend_impl)
            
            # Convert JSON-LD to VitalSigns relation objects
            relations = self._jsonld_document_to_relations(request)
            
            # Create processor
            create_processor = KGRelationsCreateProcessor(backend)
            
            # Map operation mode
            if operation_mode == OperationMode.CREATE:
                mode = CreateOperationMode.CREATE
            elif operation_mode == OperationMode.UPDATE:
                mode = CreateOperationMode.UPDATE
            else:
                mode = CreateOperationMode.UPSERT
            
            # Process relations using processor
            return await create_processor.create_or_update_relations(
                space_id, graph_id, relations, mode
            )
            
        except HTTPException:
            raise
        except Exception as e:
            self.logger.error(f"Error processing KG Relations: {e}")
            raise HTTPException(
                status_code=500,
                detail=f"Failed to process KG Relations: {str(e)}"
            )
    
    async def _delete_relations(self, space_id: str, graph_id: str, request: RelationDeleteRequest, current_user: Dict) -> RelationDeleteResponse:
        """Delete KG Relations by URIs."""
        try:
            self.logger.info(f"Deleting KG Relations in space {space_id}, graph {graph_id}: {request.relation_uris}")
            
            # Get backend implementation
            space_record = self.space_manager.get_space(space_id)
            if not space_record:
                raise HTTPException(status_code=500, detail=f"Space {space_id} not available")
            
            space_impl = space_record.space_impl
            backend_impl = space_impl.get_db_space_impl()
            if not backend_impl:
                raise HTTPException(status_code=500, detail="Backend implementation not available")
            
            # Create backend adapter
            backend = FusekiPostgreSQLBackendAdapter(backend_impl)
            
            # Create delete processor
            delete_processor = KGRelationsDeleteProcessor(backend)
            
            # Delete relations using processor
            return await delete_processor.delete_relations(space_id, graph_id, request.relation_uris)
            
        except HTTPException:
            raise
        except Exception as e:
            self.logger.error(f"Error deleting KG Relations: {e}")
            raise HTTPException(
                status_code=500,
                detail=f"Failed to delete KG Relations: {str(e)}"
            )
    
    async def _query_relations(self, space_id: str, graph_id: str, query_request: RelationQueryRequest, current_user: Dict) -> RelationQueryResponse:
        """Query KG Relations using complex criteria."""
        try:
            self.logger.info(f"Querying KG Relations in space {space_id}, graph {graph_id} with criteria: {query_request.criteria}")
            
            # Get backend implementation via generic interface
            space_record = self.space_manager.get_space(space_id)
            if not space_record:
                raise HTTPException(status_code=500, detail=f"Space {space_id} not available - server configuration error")
            
            space_impl = space_record.space_impl
            backend = space_impl.get_db_space_impl()
            if not backend:
                raise HTTPException(status_code=500, detail="Backend implementation not available")
            
            # Build SPARQL query from criteria
            sparql_query = self._build_query_relations_sparql(
                graph_id, query_request.criteria, query_request.page_size, query_request.offset
            )
            
            # Execute query via backend interface
            results = await backend.execute_sparql_query(space_id, sparql_query)
            
            # Extract relation URIs from results
            relation_uris = []
            if results and results.get("bindings"):
                for binding in results["bindings"]:
                    relation_uri = binding.get("relation", {}).get("value", "")
                    if relation_uri and relation_uri not in relation_uris:
                        relation_uris.append(relation_uri)
            
            # For now, use the actual count as total_count
            total_count = len(relation_uris)
            
            return RelationQueryResponse(
                relation_uris=relation_uris,
                total_count=total_count,
                page_size=query_request.page_size,
                offset=query_request.offset
            )
            
        except HTTPException:
            raise
        except Exception as e:
            self.logger.error(f"Error querying KG Relations: {e}")
            raise HTTPException(
                status_code=500,
                detail=f"Failed to query KG Relations: {str(e)}"
            )
    
    # Helper methods (placeholder implementations - to be completed based on mock patterns)
    
    def _build_list_relations_query(self, graph_id: str, entity_source_uri: Optional[str],
                                  entity_destination_uri: Optional[str], relation_type_uri: Optional[str],
                                  direction: str, page_size: int, offset: int) -> str:
        """Build SPARQL query for listing relations."""
        # TODO: Implement based on mock implementation patterns
        return f"""
        PREFIX haley: <{self.haley_prefix}>
        PREFIX vital: <{self.vital_prefix}>
        
        SELECT ?relation ?source ?destination ?type WHERE {{
            GRAPH <{graph_id}> {{
                ?relation a haley:Edge_hasKGRelation ;
                         vital:hasEdgeSource ?source ;
                         vital:hasEdgeDestination ?destination .
                OPTIONAL {{ ?relation a ?type }}
            }}
        }}
        LIMIT {page_size}
        OFFSET {offset}
        """
    
    def _build_count_relations_query(self, graph_id: str, entity_source_uri: Optional[str],
                                   entity_destination_uri: Optional[str], relation_type_uri: Optional[str],
                                   direction: str) -> str:
        """Build SPARQL count query for relations."""
        # TODO: Implement based on mock implementation patterns
        return f"""
        PREFIX haley: <{self.haley_prefix}>
        PREFIX vital: <{self.vital_prefix}>
        
        SELECT (COUNT(?relation) as ?count) WHERE {{
            GRAPH <{graph_id}> {{
                ?relation a haley:Edge_hasKGRelation ;
                         vital:hasEdgeSource ?source ;
                         vital:hasEdgeDestination ?destination .
            }}
        }}
        """
    
    def _build_get_relation_query(self, graph_id: str, relation_uri: str) -> str:
        """Build SPARQL query for getting specific relation."""
        return f"""
        PREFIX haley: <{self.haley_prefix}>
        PREFIX vital: <{self.vital_prefix}>
        
        SELECT ?relation ?source ?destination ?type WHERE {{
            GRAPH <{graph_id}> {{
                <{relation_uri}> a haley:Edge_hasKGRelation ;
                                vital:hasEdgeSource ?source ;
                                vital:hasEdgeDestination ?destination .
                OPTIONAL {{ <{relation_uri}> a ?type }}
                BIND(<{relation_uri}> as ?relation)
            }}
        }}
        """
    
    def _build_delete_relation_query(self, graph_id: str, relation_uri: str) -> str:
        """Build SPARQL DELETE query for specific relation."""
        return f"""
        PREFIX haley: <{self.haley_prefix}>
        PREFIX vital: <{self.vital_prefix}>
        
        DELETE {{
            GRAPH <{graph_id}> {{
                <{relation_uri}> ?p ?o .
            }}
        }}
        WHERE {{
            GRAPH <{graph_id}> {{
                <{relation_uri}> ?p ?o .
            }}
        }}
        """
    
    def _build_query_relations_sparql(self, graph_id: str, criteria, page_size: int, offset: int) -> str:
        """Build SPARQL query from relation query criteria."""
        # TODO: Implement based on criteria parameters
        return f"""
        PREFIX haley: <{self.haley_prefix}>
        PREFIX vital: <{self.vital_prefix}>
        
        SELECT ?relation WHERE {{
            GRAPH <{graph_id}> {{
                ?relation a haley:Edge_hasKGRelation ;
                         vital:hasEdgeSource ?source ;
                         vital:hasEdgeDestination ?destination .
            }}
        }}
        LIMIT {page_size}
        OFFSET {offset}
        """
    
    def _sparql_results_to_relations(self, sparql_result: Dict[str, Any]) -> List:
        """Convert SPARQL results to VitalSigns relation objects."""
        # TODO: Implement conversion to VitalSigns Edge_hasKGRelation objects
        return []
    
    def _relations_to_jsonld_document(self, relations: List):
        """Convert relation objects to JSON-LD document or object."""
        try:
            if not relations:
                return JsonLdDocument(**{
                    "@context": {"@vocab": "http://vital.ai/ontology/"},
                    "@graph": []
                })
            
            # Convert VitalSigns objects to JSON-LD
            from vital_ai_vitalsigns.model.GraphObject import GraphObject
            from vitalgraph.model.jsonld_model import JsonLdObject
            
            # Handle both list of GraphObjects and single GraphObject
            if isinstance(relations, list):
                jsonld = GraphObject.to_jsonld_list(relations)
            else:
                jsonld = GraphObject.to_jsonld_list([relations])
            
            # Ensure @type fields are present
            if '@graph' in jsonld:
                for obj in jsonld['@graph']:
                    if '@type' not in obj and 'type' in obj:
                        obj['@type'] = obj.pop('type')
            
            # Return JsonLdObject for single relation, JsonLdDocument for multiple
            if '@graph' in jsonld and len(jsonld['@graph']) == 1:
                # Single object - return JsonLdObject
                single_obj = jsonld['@graph'][0]
                if '@context' in jsonld:
                    single_obj['@context'] = jsonld['@context']
                return JsonLdObject(**single_obj)
            else:
                # Multiple objects - return JsonLdDocument
                return JsonLdDocument(**jsonld)
            
        except Exception as e:
            self.logger.error(f"Error converting relations to JSON-LD: {e}")
            return JsonLdDocument(**{
                "@context": {"@vocab": "http://vital.ai/ontology/"},
                "@graph": []
            })
    
    def _jsonld_document_to_relations(self, document: JsonLdRequest) -> List:
        """Convert JSON-LD document or object to VitalSigns relation objects."""
        try:
            # Convert Pydantic model to dict
            from vital_ai_vitalsigns.vitalsigns import VitalSigns
            vs = VitalSigns()
            
            jsonld_data = document.model_dump(by_alias=True)
            
            # If it's a JsonLdObject (single object), wrap it in a document structure
            if isinstance(document, JsonLdObject):
                # Single object - wrap in @graph array for VitalSigns
                context = jsonld_data.pop('@context', {})
                jsonld_document = {
                    '@context': context,
                    '@graph': [jsonld_data]
                }
            else:
                # Already a document with @graph
                jsonld_document = jsonld_data
            
            # Convert to VitalSigns objects using from_jsonld_list
            graph_objects = vs.from_jsonld_list(jsonld_document)
            
            if not graph_objects:
                self.logger.warning("No objects converted from JSON-LD")
                return []
            
            # Ensure we have a list
            if not isinstance(graph_objects, list):
                graph_objects = [graph_objects]
            
            # Filter for relation objects (Edge_hasKGRelation)
            from ai_haley_kg_domain.model.Edge_hasKGRelation import Edge_hasKGRelation
            filtered_relations = [obj for obj in graph_objects if isinstance(obj, Edge_hasKGRelation)]
            
            self.logger.info(f"Converted JSON-LD to {len(filtered_relations)} relation objects")
            return filtered_relations
            
        except Exception as e:
            self.logger.error(f"Error converting JSON-LD to relations: {e}")
            import traceback
            self.logger.error(traceback.format_exc())
            return []
    
    def _validate_relations(self, relations: List) -> None:
        """Validate relation objects."""
        # TODO: Implement validation logic
        pass
    
    async def _store_relations_in_space(self, space_id: str, graph_id: str, relations: List) -> List[str]:
        """Store relations in space and return created URIs."""
        if not relations:
            return []
        
        # Get space implementation
        space_record = self.space_manager.get_space(space_id)
        if not space_record:
            return []
        
        space_impl = space_record.space_impl
        backend = space_impl.get_db_space_impl()
        if not backend:
            return []
        
        # Convert relations to RDF quads
        from rdflib import URIRef
        quads = []
        created_uris = []
        
        for relation in relations:
            # Get relation triples using VitalSigns
            relation_triples = relation.to_rdf()
            
            # Convert to quads with graph URI
            graph_uri = URIRef(graph_id)
            for s, p, o in relation_triples:
                quads.append((s, p, o, graph_uri))
            
            # Track created URI
            if hasattr(relation, 'URI'):
                created_uris.append(relation.URI)
        
        # Store quads via backend
        if quads:
            await backend.add_rdf_quads_batch(space_id, quads)
        
        return created_uris
    
    async def _update_relations_in_space(self, space_id: str, graph_id: str, relations: List) -> List[str]:
        """Update relations in space and return updated URIs."""
        if not relations:
            return []
        
        # Get space implementation
        space_record = self.space_manager.get_space(space_id)
        if not space_record:
            return []
        
        space_impl = space_record.space_impl
        backend = space_impl.get_db_space_impl()
        if not backend:
            return []
        
        updated_uris = []
        
        for relation in relations:
            if not hasattr(relation, 'URI'):
                continue
            
            relation_uri = relation.URI
            
            # Delete existing relation triples
            delete_query = self._build_delete_relation_query(graph_id, relation_uri)
            await backend.execute_sparql_update(space_id, delete_query)
            
            # Add updated relation triples
            from rdflib import URIRef
            relation_triples = relation.to_rdf()
            graph_uri = URIRef(graph_id)
            quads = [(s, p, o, graph_uri) for s, p, o in relation_triples]
            
            if quads:
                await backend.add_rdf_quads_batch(space_id, quads)
                updated_uris.append(relation_uri)
        
        return updated_uris
    
    async def _upsert_relations_in_space(self, space_id: str, graph_id: str, relations: List) -> List[str]:
        """Upsert relations in space and return upserted URIs."""
        if not relations:
            return []
        
        # Get space implementation
        space_record = self.space_manager.get_space(space_id)
        if not space_record:
            return []
        
        space_impl = space_record.space_impl
        backend = space_impl.get_db_space_impl()
        if not backend:
            return []
        
        upserted_uris = []
        
        for relation in relations:
            if not hasattr(relation, 'URI'):
                continue
            
            relation_uri = relation.URI
            
            # Check if relation exists
            check_query = f"""
            ASK {{
                GRAPH <{graph_id}> {{
                    <{relation_uri}> ?p ?o .
                }}
            }}
            """
            
            result = await backend.execute_sparql_query(space_id, check_query)
            exists = result.get('boolean', False) if isinstance(result, dict) else False
            
            # Delete if exists
            if exists:
                delete_query = self._build_delete_relation_query(graph_id, relation_uri)
                await backend.execute_sparql_update(space_id, delete_query)
            
            # Add relation triples
            from rdflib import URIRef
            relation_triples = relation.to_rdf()
            graph_uri = URIRef(graph_id)
            quads = [(s, p, o, graph_uri) for s, p, o in relation_triples]
            
            if quads:
                await backend.add_rdf_quads_batch(space_id, quads)
                upserted_uris.append(relation_uri)
        
        return upserted_uris
    
    def _extract_count_from_results(self, count_results: Dict[str, Any]) -> int:
        """Extract count from SPARQL count query results."""
        if count_results and count_results.get("bindings"):
            for binding in count_results["bindings"]:
                count_value = binding.get("count", {}).get("value", "0")
                return int(count_value)
        return 0


def create_kgrelations_router(space_manager, auth_dependency) -> APIRouter:
    """Create and return the KG relations router."""
    endpoint = KGRelationsEndpoint(space_manager, auth_dependency)
    return endpoint.router